package org.w3.yourapp

import org.w3.yourapp.util.ResourceManager.fromClasspath

import javax.servlet._
import javax.servlet.http._
import unfiltered.request._
import unfiltered.response._
import unfiltered.scalate._
import unfiltered.jetty._

import org.fusesource.scalate.TemplateEngine
import org.slf4j.{Logger, LoggerFactory}

import org.w3.rdf._
import org.w3.rdb.RDB._
import org.w3.sql
import org.w3.rdf.turtle.TurtleModule
import org.w3.directmapping.DirectMappingModule
import org.w3.rdf.jena._

import org.w3.rdf2rdf._

// holds some Unfiltered plans
object YourProjectApp extends JenaModel with RDF2RDFModule with DirectMappingModule {
  
  val engine = {
    val templates = fromClasspath("templates/")
    println("Template directory extracted at " + templates.getAbsolutePath)
    val templateDirs = List(templates)
    val scalateMode = "development"
    new TemplateEngine(templateDirs, scalateMode)
  }

  val services = unfiltered.filter.Planify {
    case Path(Seg("ddl" :: ddl :: Nil)) & GET(_) =>
      try {
        Ok ~> ContentType("application/json") ~> ResponseString(QueryManager.embeddedDDLs(ddl))
      } catch {
        case e:Throwable =>  Ok ~> PlainTextContent ~> ResponseString(e.getMessage)
      }
    case Path(Seg("ddls" :: Nil)) & Params(params) =>
      try {
        val current = params("q").headOption getOrElse ""
        val limit = params("limit").headOption map (_.toInt) getOrElse 10
        Ok ~> PlainTextContent ~> ResponseString(QueryManager.embeddedDDLs.keys filter { _ contains current } take limit map (s => s+"|"+s) mkString "\n")
      } catch {
        case e:Throwable =>  Ok ~> PlainTextContent ~> ResponseString(e.getMessage)
      }
    case Path(Seg("construct" :: construct :: Nil)) & GET(_) =>
      try {
        Ok ~> ContentType("application/json") ~> ResponseString(QueryManager.embeddedCONSTRUCTs(construct))
      } catch {
        case e:Throwable =>  Ok ~> PlainTextContent ~> ResponseString(e.getMessage)
      }
    case Path(Seg("constructs" :: Nil)) & Params(params) & GET(_) =>
      try {
        val current = params("q").headOption getOrElse ""
        val limit = params("limit").headOption map (_.toInt) getOrElse 10
        Ok ~> PlainTextContent ~> ResponseString(QueryManager.embeddedCONSTRUCTs.keys filter { _ contains current } take limit map (s => s+"|"+s) mkString "\n")
      } catch {
        case e:Throwable =>  Ok ~> PlainTextContent ~> ResponseString(e.getMessage)
      }
    case Path(Seg("parse-ddl" :: Nil)) & Params(params) & POST(_) =>
      try {
        val SQLParser = sql.SqlParser()
        val db = SQLParser.toDB(params("ddl").headOption getOrElse "")
        Ok ~> PlainTextContent ~> ResponseString("DDL successfully parsed")
      } catch {
        case e:Throwable =>  Ok ~> PlainTextContent ~> ResponseString(e.getMessage)
      }
    case Path(Seg("parse-construct" :: Nil)) & Params(params) & POST(_) =>
      try {
        val constructs = Construct.extractAll(params("construct").headOption getOrElse "")
        Ok ~> PlainTextContent ~> ResponseString("SPARQL CONSTRUCT successfully parsed")
      } catch {
        case e:Throwable =>  Ok ~> PlainTextContent ~> ResponseString(e.getMessage)
      }
    case Path(Seg("dm-then-construct" :: Nil)) & Params(params) & POST(_) =>
      try {
        val ddl = params("ddl").headOption getOrElse ""
        val construct = params("construct").headOption getOrElse ""
        Ok ~> PlainTextContent ~> ResponseString(dmThenConstruct(ddl, construct))
      } catch {
        case e:Throwable =>  Ok ~> PlainTextContent ~> ResponseString(e.getMessage)
      }
  }
    
  private def dmThenConstruct(ddl:String, constructs:String):String = {
    import DirectMapping._
    val SQLParser = sql.SqlParser()
    val db = SQLParser.toDB(ddl)
    DirectMapping.MinEncode = false
    val computedGraph:Graph = databaseSemantics(db)
    val allConstructs:List[Construct] = Construct.extractAll(constructs)
    val resultGraph = RDF2RDF(computedGraph) | allConstructs
    this.toString(resultGraph)
  }


  val index = unfiltered.filter.Planify {
    case req @ Path(Seg(Nil)) & Params(params) => {
      import DirectMapping._
      val SQLParser = sql.SqlParser()
      val defaultDDL = "emp_addr"
      val defaultCONSTRUCTs = "emp_addr2FOAF"
      val (ddlName, ddl, ddlMessage) = params("ddl").headOption match {
        case Some(name) if QueryManager.embeddedDDLs contains name => {
          (name, QueryManager.embeddedDDLs(name), "DDL successfully loaded")
        }
        case Some(name) => {
          (defaultDDL, QueryManager.embeddedDDLs(defaultDDL), name + "not found: default DDL used instead")
        }
        case None =>  {
          (defaultDDL, QueryManager.embeddedDDLs(defaultDDL), "Default DDL used")
        }
      }

      val (constructName, constructs, constructMessage) = params("construct").headOption match {
        case Some(name) if QueryManager.embeddedCONSTRUCTs contains name => {
          (name, QueryManager.embeddedCONSTRUCTs(name), "SPARQL CONSTRUCT successfully loaded")
        }
        case Some(name) => {
          (defaultCONSTRUCTs, QueryManager.embeddedCONSTRUCTs(defaultCONSTRUCTs), name + "not found: default SPARQL CONSTRUCT used instead")
        }
        case None =>  {
          (defaultCONSTRUCTs, QueryManager.embeddedCONSTRUCTs(defaultCONSTRUCTs), "Default SPARQL CONSTRUCT used")
        }
      }

      Ok ~> ContentType("application/xhtml+xml") ~> Scalate(req, "index.ssp",
                                                            "ddlName" -> "",
                                                            "ddl" -> ddl,
                                                            "ddlMessage" -> ddlMessage,
                                                            "constructName" -> "",
                                                            "construct" -> constructs,
                                                            "constructMessage" -> constructMessage,
                                                            "result" -> dmThenConstruct(ddl, constructs))(engine)
    }
  }

}

import YourProjectApp._

object VirtualTrainerAppMain {

  val logger:Logger = LoggerFactory.getLogger(this.getClass)

  // regular Java main
  def main(args: Array[String]) {
    
    // can provide the port as first argument
    // default to 2719
    val port = args.toList.headOption map { _.toInt } getOrElse 2719
    
    // configures and launches a Jetty server
    unfiltered.jetty.Http(port)
    // serves static files
    .context("/scripts"){ ctx:ContextBuilder =>
      ctx.resources(fromClasspath("scripts/").toURI.toURL)
    }.context("/styles"){ ctx:ContextBuilder =>
      ctx.resources(fromClasspath("styles/").toURI.toURL)
    // Unfiltered filters
    }.filter(services).filter(index).run()
    
  }

}
