package org.w3.rdb

/**
 * Relational structure
 */
object RDB {

  /**
   * a Database defines a set of relations
   */
  case class Database(private val m:Map[RelName, Relation]) {
    // tricky and evil
    // when the Database is created, it can go through all the foreign key and set the database field
    // this way, a Target can know directly the Relation without using Database
    for {
      (_, r) <- m
      fk <- r.fks
    } {
      fk.target.db = Some(this)
    }

    def apply(rn:RelName) = m(rn)
    /** returns all the relations in this database */
    def relations:Set[Relation] = m.values.toSet
  }

  object Database {
    def apply (l:Relation*):Database =
      Database(l map { r => (r.name -> r) } toMap)
  }

  /**
   * definition of a Relation
   * pk should be seen as the function extracting the unique Primary Key from candidates (if present)
   * the current SQL parser ensures the pk is materialized twice
   */
  case class Relation(name:RelName,
		      header:Header,
		      body:List[Tuple],
		      candidates:List[CandidateKey],
		      pk:Option[CandidateKey],
		      fks:ForeignKeys) {
    var rowCounter = 0
    for {
      t <- body
    } {
      t._relation = Some(this)
      rowCounter = rowCounter + 1
      t.rowId = rowCounter
    }

    /** adds a tuple in the body of the relation */
    def +(t:Tuple):Relation = this.copy(body = body :+ t)
  }

  case class RelName(n:String) {
    override def toString = n
  }

  /**
   * a Header maps an attribute name to a SQL datatype
   * the set of attributes names can be extracted (it's really not ordered)
   */
  case class Header(private val m:Map[AttrName, Datatype]) extends PartialFunction[AttrName, Datatype] {
    def apply(a:AttrName) = m(a)
    def isDefinedAt(a:AttrName) = m isDefinedAt a
  }

  object Header {
    def apply (s:(String, Datatype)*):Header =
      Header(s map { case (name, datatype) => (AttrName(name), datatype) } toMap)
  }

  /**
   * a tuple maps an attribute name to a value in a cell
   * the tuple does not carry the information from the header so when it's needed,
   * this has to be retrieved from the relation
   * especially, the order of the attribute names is not known within the tuple
   */
  case class Tuple(private val m:Map[AttrName, CellValue],
                   var rowId:Int = scala.util.Random.nextInt())
  extends PartialFunction[AttrName, CellValue] {

    override def toString = m.toString

    // hack for the tuple to know the relation it belongs to
    var _relation:Option[Relation] = None
    lazy val relation:Relation = _relation.get

    def apply (a:AttrName) = m(a)

    def isDefinedAt(a:AttrName) = m isDefinedAt a

    /**
     * returns all the lexical values corresponding to the given as
     * the order from the list is preserved
     * we assume the values are restricted to non null cells
     * so it's safe to call only with a primary key
     */
    def lexvalues (as:AttrList /* forall a in as, a is in this Tuple */):List[LexicalValue] =
      as.attrs map {
        m(_) match {
          case lexicalValue @ LexicalValue(_) => lexicalValue
          case ␀ => error("this value MUST not be null")
        }
      }

    /**
     * returns all the lexical values corresponding to the given as
     * the order from the list is preserved
     */
    def cellvalues (as:AttrList /* forall a in as, a is in this Tuple */):List[CellValue] =
      as.attrs map { m(_) }

    /**
     * returns all foreign keys that have only not-null values in it
     */
    def references:Set[ForeignKey] = {
      val nullAttributes:Set[AttrName] =
	m collect { case (attrName, cellValue) if cellValue == ␀ => attrName } toSet

      relation.fks filter { case ForeignKey(as, _) => nullAttributes & as.toSet isEmpty  }
    }

    /**
     * returns all the not null attribute names
     * such that they don't also define a unary foreign key
     */
    def scalars:Set[AttrName] = {
      val notNullAttributes:Set[AttrName] =
	m collect { case (attrName, cellValue) if cellValue != ␀ => attrName } toSet

      notNullAttributes filterNot { attrName => relation.fks definesActuallyUnaryFK attrName }
    }

    def dereference(fk:ForeignKey):Tuple = {
      val values = this.cellvalues(fk)
      val targetRelation:Relation = fk.target.relation
      targetRelation.body find { t => values == t.cellvalues(fk.target.key) } get
    }

  }

  object Tuple {
    def apply (s:(String, CellValue)*):Tuple =
      Tuple(s map { case (name, cellValue) => (AttrName(name), cellValue) } toMap)
  }

  /**
   * a cell value has either a lexical value or the NULL value
   */
  abstract class CellValue
  case class LexicalValue(s:String) extends CellValue
  case object ␀ extends CellValue

  /**
   * ForeignKeys abstracts a set of foreign keys
   * the expected operation are available
   */
  case class ForeignKeys (private val fks:Set[ForeignKey]) {
    def definesActuallyUnaryFK(a:AttrName):Boolean = fks filter { _.isUnary } exists { _.attrs contains a }
    def filter(p: ForeignKey => Boolean):Set[ForeignKey] = fks filter p
    def foreach(f:ForeignKey => Unit) = fks foreach f
    // sparql2sql
    def contains(a:AttrList):Boolean = fks exists { fk => fk.attrs == a.attrs }
    def targetOf(as:AttrList):Target = fks collect { case ForeignKey(attrs, target) if as.attrs == attrs => target } head
  }

  object ForeignKeys {
    def apply (fks:(List[String], Target)*):ForeignKeys =
      ForeignKeys(fks map { case (keys, target) => ForeignKey(keys map { AttrName(_) }, target)} toSet)
  }

  trait AttrList {
    val attrs:List[AttrName]
    /**
     * by definition, a key is unary if it has only one attribute
     */
    def isUnary:Boolean = attrs.length == 1
    // for sparql2sql
    def apply(index:Int):AttrName = attrs(index)
  }

  object AttrList {
    def apply(a:AttrName):AttrList = new AttrList { val attrs = List(a) }
    def apply(as:List[AttrName]):AttrList = new AttrList { val attrs = as }
  }

  case class ForeignKey(attrs:List[AttrName], target:Target) extends AttrList

  case class Target(rn:RelName, key:CandidateKey) {
    // hack: this field will be updated by the Database when it will be created
    // as the target is fully know at that moment
    var db:Option[Database] = None
    lazy val relation:Relation = db.get(rn)
  }

  case class CandidateKey (attrs:List[AttrName]) extends AttrList

  object CandidateKey {
    def apply (l:String*):CandidateKey =
      CandidateKey(l map { AttrName(_) } toList)
  }

  case class AttrName(n:String) {
    override def toString = n
  }

  case class Datatype(name:String) {
    override def toString = "/* " + name + " */"
  }

  // TODO: use Enumeration so we can go through the datatypes?
  object Datatype {
    val CHAR = Datatype("Char")
    val VARCHAR = Datatype("Varchar")
    val STRING = Datatype("String")
    val INTEGER = Datatype("Int")
    val FLOAT = Datatype("Float")
    val DOUBLE = Datatype("Double")
    val DATE = Datatype("Date")
    val TIME = Datatype("Time")
    val TIMESTAMP = Datatype("Timestamp")
    val DATETIME = Datatype("Datetime")
  }

}
