package org.w3.rdf2rdf

import org.w3.rdf.jena._
import com.hp.hpl.jena.query._
import com.hp.hpl.jena.rdf.model._

object RDF2RDFModule {
  val base = "dm:"
}

case class Construct(query:Query) {
  assert(query.getQueryType == Query.QueryTypeConstruct, "You must provide a valid SPARQL 1.1 Construct query")
}

object Construct {
  def apply(query:String, base:String = RDF2RDFModule.base):Construct = {
    val construct = QueryFactory.create(query, base, Syntax.syntaxSPARQL_11)
    Construct(construct)
  }

  val pattern = """(?s)(?i)CONSTRUCT +?\{.*?\} +?WHERE +?\{.*?\}""".r

  def extractAll(query:String, base:String = RDF2RDFModule.base):List[Construct] = {
    val prefixes = query.split("\n").filter{ _.matches("""(?i)^\s*prefix.*$""")}.mkString("\n")
    def acc(lines:List[String], currentQuery:String, queries:List[String]):List[String] = {
      lines match {
        case line :: lines if line startsWith "---" => acc(lines, "", currentQuery :: queries)
        case line :: lines => acc(lines, currentQuery+"\n"+line, queries)
        case Nil => currentQuery :: queries
      }
    }
    val constructs = acc(query.split("\n").toList, "", Nil)
    constructs map { construct => Construct(prefixes+construct.replaceAll("CONSTRUCT$", ""), base) } toList
  }

}


import RDF2RDFModule._

trait RDF2RDFModule extends JenaModel {

  case class RDF2RDF(graph:Graph) {

    val model = {
      val m1 = ModelFactory.createModelForGraph(graph.jenaGraph)
      val buffer = new java.io.ByteArrayOutputStream
      m1.write(buffer, "TURTLE", base)
      // m1.write(System.out, "TURTLE", base)
      val m2 = ModelFactory.createDefaultModel()
      m2.read(new java.io.StringReader(buffer.toString("UTF-8")), base, "TURTLE")
      // m2.write(System.out, "TURTLE", null)
      m2
    }

    def | (construct:Construct):Graph = {
      val qe = QueryExecutionFactory.create(construct.query, model)
      val resultModel = qe.execConstruct
      val g1 = new Graph(resultModel.getGraph)
      g1
    }

    /**
     * the union of all the CONSTRUCTs applied in parallel against to the
     * same RDF dataset
     */
    def | (constructs:List[Construct]):Graph = {
      val graphs:List[Graph] = constructs map { construct => this | construct }
      graphs.foldLeft(Graph.empty) { _ ++ _ }
    }

  }

  object RDF2RDF {
    implicit def graph2RDF2RDF(graph:Graph):RDF2RDF = RDF2RDF(graph)
  }

  import com.hp.hpl.jena.rdf.model._

  def getGraph(s:String, base:String = null):Graph = {
    val model = ModelFactory.createDefaultModel()
    model.read(new java.io.StringReader(s), base, "TURTLE")
    return new Graph(model.getGraph)
  }

  def toString(g:Graph):String = {
    val model = ModelFactory.createModelForGraph(g.jenaGraph)
    val buffer = new java.io.ByteArrayOutputStream
    model.write(buffer, "TURTLE", base)
    val result = buffer.toString("UTF-8").replaceAll("<"+base, "<")
    result
  }

  def dumpCutingBase(g:Graph):Unit = {
    val result = toString(g)
    println(result)
  }

}


object Main {

  import com.hp.hpl.jena.rdf.model._

  def main(args: Array[String]): Unit = {

    val turtleString = """
<a> <b> <c> .
"""

    val queryString = """
CONSTRUCT {
  <otherA> ?p <otherC>
} WHERE {
  <a> ?p <c>
}
"""

    val base = "/"

    val model:Model = ModelFactory.createDefaultModel()
    model.read(new java.io.StringReader(turtleString), base, "TURTLE")
    // val s = model.createResource("a")
    // val p = model.createResource("b")
    // val o = model.createResource("c")
    // val statement = model.create
    model.write(System.out, "TURTLE", base)

    val query:Query = QueryFactory.create(queryString, base)
    val qexec:QueryExecution = QueryExecutionFactory.create(query, model)
    val resultModel:Model = qexec.execConstruct()

    resultModel.write(System.out, "TURTLE", base)

  }

}
