package org.w3.rdf2rdf

import org.scalatest.FunSuite
import org.scalatest.matchers.ShouldMatchers
import org.w3.rdf.jena._
import org.w3.directmapping._

class RDF2RDFTest extends FunSuite with ShouldMatchers with RDF2RDFModule {

  test("http://ivan-herman.name/2010/11/19/my-first-mapping-from-direct-mapping/") {

    val graph = getGraph("""
<Book/ID=0006511409X#_> a <Book> ;
  <Book#ISBN> "0006511409X" ;
  <Book#Title> "The Glass Palace" ;
  <Book#Year>  "2000" ;
  <Book#Author> <Author/ID=id_xyz#_> .

<Author/ID=id_xyz#_> a <Author> ;
  <Author#ID> "id_xyz" ;
  <Author#Name> "Ghosh, Amitav" ;
  <Author#Homepage> "http://www.amitavghosh.com" .
""", base="http://book.example/")


    val construct = Construct("""
PREFIX a: <http://ivan.book.example/>
PREFIX fn: <http://www.w3.org/2005/xpath-functions#>
CONSTRUCT {
  ?id a:title ?title ;
      a:year  ?year ;
      a:author _:x .
  _:x a:name ?name ;
      a:homepage ?hp .
}
WHERE {
  SELECT (IRI(fn:concat("http://ivan.book.example/", ?isbn)) AS ?id)
          ?title ?year ?name
         (IRI(?homepage) AS ?hp)
  {
    ?book a <Book> ;
          <Book#ISBN> ?isbn ;
          <Book#Title> ?title ;
          <Book#Year> ?year ;
          <Book#Author> ?author .
    ?author a <Author> ;
            <Author#Name> ?name ;
            <Author#Homepage> ?homepage .
  }
}""", base="http://book.example/")


    val resultGraph = RDF2RDF(graph) | construct

    val expectedGraph = getGraph("""
<0006511409X> <title> "The Glass Palace" ;
              <year>  "2000" ;
              <author> _:y .
_:y <name> "Ghosh, Amitav" ;
    <homepage> <http://www.amitavghosh.com> .
""", base="http://ivan.book.example/")

    assert(expectedGraph === resultGraph)

  }

  test("// CONSTRUCTs") {

    val graph = getGraph("""
<P:8> <P:fname> "Bob" ; <P:lname> "Smith" ; <P:addr> <A:18> .
<P:9> <P:fname> "Sue" ; <P:lname> "Jones" .
<A:18> <A:streetNo> 32 ; <A:streetName> "Vassar" .
<A:76> <A:streetNo> 77 ; <A:streetName> "Mass Ave" .
""", base="")


    val c1 = Construct("""
CONSTRUCT {
  ?who <foaf:given> ?fname ; <foaf:family> ?lname
} WHERE {
  ?who <P:fname> ?fname ; <P:lname> ?lname
}
""", base="")

    val c2 = Construct("""
CONSTRUCT {
  ?addr <vcard:streetNumber> ?no ; <vcard:streetName> ?name
} WHERE {
  ?addr <A:streetNo> ?no ; <A:streetName> ?name
}
""", base="")

    val c3 = Construct("""
CONSTRUCT {
  ?who <vcard:homeAddr> ?addr
} WHERE {
  ?who <P:addr> ?addr
}
""", base="")


    val resultGraph = RDF2RDF(graph) | List(c1, c2, c3)

//     val expectedGraph = getGraph("""
// <0006511409X> <title> "The Glass Palace" ;
//               <year>  "2000" ;
//               <author> _:y .
// _:y <name> "Ghosh, Amitav" ;
//     <homepage> <http://www.amitavghosh.com> .
// """, base="http://ivan.book.example/")

//     dumpCutingBase(resultGraph)

  }

  test("// CONSTRUCTs parsed") {

    val graph = getGraph("""
<P:8> <P:fname> "Bob" ; <P:lname> "Smith" ; <P:addr> <A:18> .
<P:9> <P:fname> "Sue" ; <P:lname> "Jones" .
<A:18> <A:streetNo> 32 ; <A:streetName> "Vassar" .
<A:76> <A:streetNo> 77 ; <A:streetName> "Mass Ave" .
""", base="")


    val constructs = Construct.extractAll("""
CONSTRUCT {
  ?who <foaf:given> ?fname ; <foaf:family> ?lname
} WHERE {
  ?who <P:fname> ?fname ; <P:lname> ?lname
}
---
CONSTRUCT {
  ?addr <vcard:streetNumber> ?no ; <vcard:streetName> ?name
} WHERE {
  ?addr <A:streetNo> ?no ; <A:streetName> ?name
}
---
CONSTRUCT {
  ?who <vcard:homeAddr> ?addr
} WHERE {
  ?who <P:addr> ?addr
}
""", base="")


    val resultGraph = RDF2RDF(graph) | constructs

    println(resultGraph)

  }
}


import org.w3.rdb.RDB._
import org.w3.sql
import org.w3.rdf.turtle.TurtleModule

class RDB2RDF2RDFTest extends FunSuite with ShouldMatchers with RDF2RDFModule with DirectMappingModule with TurtleModule {

  import DirectMapping._

  val SqlParser = sql.SqlParser()

  val turtleParser = new TurtleParser { }

  test("http://lists.w3.org/Archives/Public/public-rdb2rdf-wg/2010Nov/0003.html") {

    val db = SqlParser.toDB("""
CREATE TABLE Person (ID INT PRIMARY KEY, fname STRING);
CREATE TABLE Address (ID INT PRIMARY KEY, city STRING);
CREATE TABLE PersonAddress (personID INT PRIMARY KEY,
                            addressID INT PRIMARY KEY,
                            FOREIGN KEY (personID) REFERENCES Person(ID),
                            FOREIGN KEY (addressID) REFERENCES Address(ID));
INSERT INTO Person (ID, fname) VALUES (7, "Bob");
INSERT INTO Person (ID, fname) VALUES (8, "Sue");
INSERT INTO Address (ID, city) VALUES (18, "Cambridge");
INSERT INTO Address (ID, city) VALUES (19, "Austin");
INSERT INTO PersonAddress (personID, addressID) VALUES (7, 18);
INSERT INTO PersonAddress (personID, addressID) VALUES (7, 19);
INSERT INTO PersonAddress (personID, addressID) VALUES (8, 19);
""")

    val graph = databaseSemantics(db)

    val construct = Construct("""
CONSTRUCT {
  ?s <p> ?o
} WHERE {
  ?s <Address#ID> ?o
}
""")

    println(construct)

    val resultGraph = RDF2RDF(graph) | construct

    dumpCutingBase(resultGraph)

  }

}

