package org.w3.sparql

import scala.util.parsing.combinator._
import java.net.URI

object rdf {

  sealed abstract class Literal(val lexicalForm:String)
  case class PlainLiteral(override val lexicalForm:String, langtag:Option[LangTag]) extends Literal(lexicalForm) {
    override def toString = "\"" + lexicalForm + "\"" + { if (langtag.isDefined) langtag.get }
  }
  case class TypedLiteral(override val lexicalForm:String, datatype:IRI) extends Literal(lexicalForm) {
    override def toString = "\"" + lexicalForm + "\"^^" + datatype
  }
  case class LangTag(s:String)

  case class IRI(iri:String) { override def toString = '"' + iri + '"' }
  lazy val StringDatatype = IRI("http://www.w3.org/2001/XMLSchema#string")
  lazy val IntegerDatatype = IRI("http://www.w3.org/2001/XMLSchema#integer")
  lazy val DateDatatype = IRI("http://www.w3.org/2001/XMLSchema#date")
  lazy val DateTimeDatatype = IRI("http://www.w3.org/2001/XMLSchema#dateTime")

}

object MyParsers extends RegexParsers {

  val uri = """[a-zA-Z0-9:/#_\.\-\+]+""".r
  val integer = """[0-9]+""".r
  val name = """[a-zA-Z][a-zA-Z0-9_-]*|[a-zA-Z_][a-zA-Z0-9_]+""".r
  var prefixes:Map[String, String] = Map()
  var bnodes:Map[String, BNode] = Map()
  var nextBNode = 1
}

import MyParsers._

case class Select(distinct:Boolean, attrs:SparqlAttributeList, gp:GraphPattern, order:List[OrderElt], offset:Option[Int], limit:Option[Int]) {
  override def toString =
    "SELECT "+
    { if (distinct) "DISTINCT " else "" }+
    attrs+"\n"+gp+" "+
    { if (order.size > 0) {order.map(o => o.toString).mkString("ORDER BY ", " ", " ") } else "" }+
    { if (offset.isDefined) {"OFFSET " + offset.get + " "} else "" }+
    { if (limit.isDefined) {"LIMIT " + limit.get + " "} else "" }
}
case class Construct(head:TriplesBlock, gp:GraphPattern)
case class SparqlAttributeList(attributelist:List[Var]) {
  override def toString = attributelist.toList.sortWith((l, r) => l.s < r.s).mkString(" ")
}

sealed abstract class GraphPattern {
  def findVars ():Set[Var] = {
    this match {
      case TableFilter(gp2:GraphPattern, expr:Expression) =>
	gp2.findVars

      case TriplesBlock(triplepatterns) =>
	/* Examine each triple, updating the compilation state. */
	triplepatterns.foldLeft(Set[Var]())((x, y) => x ++ y.findVars)

      case TableConjunction(list) =>
	/* Examine each triple, updating the compilation state. */
	list.foldLeft(Set[Var]())((x, y) => x ++ y.findVars)

      case OptionalGraphPattern(gp2) =>
	/* Examine each triple, updating the compilation state. */
	gp2.findVars

      case x => error("no code to handle " + x)
    }
  }
  def findAssignables ():Set[Assignable] = {
    this match {
      case TableFilter(gp2:GraphPattern, expr:Expression) =>
	gp2.findAssignables

      case TriplesBlock(triplepatterns) =>
	/* Examine each triple, updating the compilation state. */
	triplepatterns.foldLeft(Set[Assignable]())((x, y) => x ++ y.findAssignables)

      case TableConjunction(list) =>
	/* Examine each triple, updating the compilation state. */
	list.foldLeft(Set[Assignable]())((x, y) => x ++ y.findAssignables)

      case OptionalGraphPattern(gp2) =>
	/* Examine each triple, updating the compilation state. */
	gp2.findAssignables

      case x => error("no code to handle " + x)
    }
  }

  def trim (terms:Set[Term]):GraphPattern = {
    this match {
      case TableFilter(gp2:GraphPattern, expr:Expression) =>
	TableFilter(gp2.trim(terms), expr)

      case TriplesBlock(triplepatterns) => {
	val r0 = new org.w3.util.GraphAnalyzer[Term](Map[Term, Set[Term]]())
	/* Examine each triple, updating the compilation state. */
	val r = triplepatterns.foldLeft(r0)((r, triple) => r.pair(triple.s, triple.o))
	val useful = triplepatterns.foldLeft(Set[TriplePattern]())((s, t) => {
	  if (r.neededFor(terms, t.s, Set(t.o)) &&
	      r.neededFor(terms, t.o, Set(t.s))) s + t
	  else s
	})
	val useful2 =
	  if (useful.size == 0)
	    triplepatterns.foldLeft(Set[TriplePattern]())((s, t) => {
	      if (r.neededFor(terms, t.s, Set(t.o)) ||
		  r.neededFor(terms, t.o, Set(t.s))) s + t
	      else s
	    })
          else useful
	TriplesBlock(useful2.toList)
      }

      case TableConjunction(list) =>
	/* Examine each triple, updating the compilation state. */
	TableConjunction(list.map(gp2 => gp2.trim(terms)))

      case OptionalGraphPattern(gp2) =>
	/* Examine each triple, updating the compilation state. */
	OptionalGraphPattern(gp2.trim(terms))

      case x => error("no code to handle " + x)
    }
  }

  def simplify ():GraphPattern = {
    this match {
      case TableFilter(gp2:GraphPattern, expr:Expression) =>
	TableFilter(gp2.simplify, expr)

      case tb:TriplesBlock => tb

      case TableConjunction(list) => {
	/* Eliminate series of TriplesBlocks. */
	val (conjuncts, triples) = list.foldLeft((List[GraphPattern](), List[TriplePattern]()))((pair, gp2) => {
	  val (conj, trip) = pair
	  val gp3 = gp2.simplify
	  gp3 match {
	    case TriplesBlock(triplepatterns) =>
	      (conj, trip ++ triplepatterns)
	    case x => {
	      (conj ++ List(TriplesBlock(trip)), List[TriplePattern]())
	    }
	  }
	})
	val conj2 =
	  if (triples.size > 0) conjuncts ++ List(TriplesBlock(triples))
	  else conjuncts
	if (conj2.size > 1) TableConjunction(conj2)
	else if (conj2.size == 1) conj2(0)
	else TriplesBlock(List[TriplePattern]())
      }

      case OptionalGraphPattern(gp2) =>
	/* Examine each triple, updating the compilation state. */
	OptionalGraphPattern(gp2.simplify)

      case x => error("no code to handle " + x)
    }
  }
}

case class TriplesBlock(triplepatterns:List[TriplePattern]) extends GraphPattern {
  override def toString = "{\n  " + (triplepatterns.toList.map(s => s.toString.replace("\n", "\n  ")).mkString("", " .\n  ", " .\n")) + "}"
  override def equals (other:Any):Boolean = other match {
    case that:TriplesBlock => (that canEqual this) && triplepatterns.toSet == that.triplepatterns.toSet
    case _ => false
  }
  override def canEqual(other : Any) : Boolean = other.isInstanceOf[TriplesBlock]
  override def hashCode:Int = 41*triplepatterns.toSet.hashCode
}
case class TableConjunction(gps:List[GraphPattern]) extends GraphPattern {
  assert (!(gps exists (x => { x match { case TableConjunction(_) => true case _ => false } })))
  override def toString = "{\n  " + (gps.toList.map(s => s.toString.replace("\n", "\n  ")).mkString("\n  ")) + "\n}\n"
}
case class TableDisjunction(gps:List[GraphPattern]) extends GraphPattern {
  override def toString = "{\n  " + (gps.toList.map(s => s.toString.replace("\n", "\n  ")).mkString("\nUNION\n  ")) + "\n}\n"
}
case class TableFilter(gp:GraphPattern, expr:Expression) extends GraphPattern {
  override def toString = gp.toString + "\nFILTER (" + expr.toString + ")"
}
case class OptionalGraphPattern(gp:GraphPattern) extends GraphPattern {
  override def toString = "OPTIONAL " + gp.toString
}
case class MinusGraphPattern(gp:GraphPattern) extends GraphPattern {
  override def toString = "MINUS " + gp.toString
}
case class GraphGraphPattern(gp:GraphPattern) extends GraphPattern

case class TriplePattern(s:Term, p:Term, o:Term) {
  override def toString = s + " " + p + " " + o
  def findVars ():Set[Var] = {
    val varS:Set[Var] = s match {
      case TermVar(v) => Set(v)
      case _                 => Set()
    }
    val varO:Set[Var] = o match {
      case TermVar(v) => Set(v)
      case _                 => Set()
    }
    varS ++ varO
  }
  def findAssignables ():Set[Assignable] = {
    val varS:Set[Assignable] = s match {
      case TermVar(v) => Set(VarAssignable(v))
      case TermBNode(v) => Set(BNodeAssignable(v))
      case _                 => Set()
    }
    val varO:Set[Assignable] = o match {
      case TermVar(v) => Set(VarAssignable(v))
      case TermBNode(b) => Set(BNodeAssignable(b))
      case _                 => Set()
    }
    varS ++ varO
  }
}

case class Literal(lit:rdf.Literal) {
  override def toString = lit match {
    case rdf.TypedLiteral(s, rdf.IntegerDatatype) => s
    case _ => lit.toString
  }
}

sealed abstract class Assignable { val s:String }

case class VarAssignable(v:Var) extends Assignable { val s = v.s }
case class BNodeAssignable(b:BNode) extends Assignable { val s = b.s }

case class Var(s:String) {
  override def toString = "?" + s
}

case class Uri(s:String) {
  override def toString = "<" + s + ">"
}

case class BNode(s:String) {
  override def toString = "_:" + s
}

case class Expression(conjuncts:List[PrimaryExpression]) {
  override def toString = conjuncts.map(e => e.toString()).mkString(" && ")
}

sealed abstract class PrimaryExpression
case class PrimaryExpressionEq(left:SparqlTermExpression, right:SparqlTermExpression) extends PrimaryExpression {
  override def toString = left.toString() + " = " + right.toString
}
case class PrimaryExpressionLt(left:SparqlTermExpression, right:SparqlTermExpression) extends PrimaryExpression {
  override def toString = left.toString() + " < " + right.toString
}
case class PrimaryExpressionGt(left:SparqlTermExpression, right:SparqlTermExpression) extends PrimaryExpression {
  override def toString = left.toString() + " > " + right.toString
}
case class SparqlTermExpression(term:Term) extends PrimaryExpression {
  override def toString = term.toString
}

case class OrderElt(desc:Boolean, expr:Expression) {
  override def toString = { if (desc) "DESC" else "ASC" } + "(" + expr.toString + ")"
}

sealed trait Term
case class TermUri(u:Uri) extends Term {
  override def toString = u.toString
}
case class TermBNode(b:BNode) extends Term {
  override def toString = b.toString
}
case class TermVar(v:Var) extends Term {
  override def toString = v.toString
}
case class TermLit(lit:Literal) extends Term {
  override def toString = lit.toString
}


case class Sparql() extends JavaTokenParsers {

  def select:Parser[Select] =
    rep(prefixdecls) ~ "SELECT" ~ opt("DISTINCT") ~ attributelist ~ opt("WHERE") ~ groupgraphpattern ~ opt(order) ~ opt(offset) ~ opt(limit) ^^ {
      case x~"SELECT"~d~a~w~gp~ord~off~lim => Select(d.isDefined, a, gp, ord.getOrElse(List[OrderElt]()), off, lim)
    }

  def order:Parser[List[OrderElt]] = 
    "ORDER" ~ "BY" ~ rep(orderelt) ^^ { case o~b~elts => elts }

  def orderelt:Parser[OrderElt] = (
      "ASC" ~ "(" ~ expression ~ ")" ^^ { case a~o~expr~c => OrderElt(false, expr) }
    | "DESC" ~ "(" ~ expression ~ ")" ^^ { case a~o~expr~c => OrderElt(true, expr) }
    | varr ^^ { case v => OrderElt(false, Expression(List(SparqlTermExpression(TermVar(v))))) }
  )
    
  def offset:Parser[Int] =
    "OFFSET" ~ integer ^^ { case o~i => i.toInt }

  def limit:Parser[Int] =
    "LIMIT" ~ integer ^^ { case o~i => i.toInt }

  def construct:Parser[Construct] =
    rep(prefixdecls) ~ "CONSTRUCT" ~ constructpattern ~ opt("WHERE") ~ groupgraphpattern ^^ { case x~"CONSTRUCT"~a~w~gp => Construct(a, gp) }

  def constructpattern:Parser[TriplesBlock] =
    "{" ~ opt(triplesblock) ~ "}" ^^ { case "{"~tbOPT~"}" => tbOPT.getOrElse(TriplesBlock(List[TriplePattern]())) }

  def prefixdecls:Parser[Unit] =
    "PREFIX" ~ name ~ ":" ~ qnameORuri ^^ { case "PREFIX"~pre~":"~u => prefixes += (pre -> u.s) }

  def filter:Parser[Expression] =
    "FILTER" ~ "(" ~ expression ~ ")" ^^ { case "FILTER"~"("~expression~")" => expression }

  def expression:Parser[Expression] = 
    repsep(primaryexpression, "&&") ^^ 
    { Expression(_) }

  def primaryexpression:Parser[PrimaryExpression] = (
      value ~ "=" ~ value ^^
      { case left ~ "=" ~ right => PrimaryExpressionEq(left, right) }
    | value ~ "<" ~ value ^^
      { case left ~ "<" ~ right => PrimaryExpressionLt(left, right) }
    | value ~ ">" ~ value ^^
      { case left ~ ">" ~ right => PrimaryExpressionGt(left, right) }
    | value
  )

  def value:Parser[SparqlTermExpression] = (
      qnameORuri ^^ { case x => SparqlTermExpression(TermUri(x)) }
    | varr ^^ { x => SparqlTermExpression(TermVar(x)) }
    | literal ^^ { x => SparqlTermExpression(TermLit(x)) }
  )

  def attributelist:Parser[SparqlAttributeList] =
    rep(varr) ^^ { SparqlAttributeList(_) }

  def groupgraphpattern:Parser[GraphPattern] = (
    "{" ~ opt(triplesblock) ~ rep(graphpatternnottriplesORfilter ~ opt(triplesblock)) ~ "}" ^^
    {
      case "{"~tbOPT~gpntORf_tbOPT~"}" => {

// case class TriplesBlock(triplepatterns:List[TriplePattern]) extends GraphPattern
// case class TableConjunction(gps:List[GraphPattern]) extends GraphPattern
// case class TableDisjunction(gps:List[GraphPattern]) extends GraphPattern
// case class TableFilter(gp:GraphPattern, expr:Expression) extends GraphPattern
// case class OptionalGraphPattern(gp:GraphPattern) extends GraphPattern
// case class GraphGraphPattern(gp:GraphPattern) extends GraphPattern

	// println("groupgraphpattern(" + tbOPT + ", " + gpntORf_tbOPT + ")")
	val init:Option[GraphPattern] = tbOPT
	gpntORf_tbOPT.foldLeft(init)((gp, lentry) => {//println("match: " + (gp, lentry))
	  // print("case (" + gp + ", " + lentry + ")")
	  (gp, lentry) match {
	    case (Some(TableFilter(TriplesBlock(l), Expression(lexp))), ~(TableFilter(null, Expression(expr)), Some(TriplesBlock(r)))) => Some(TableFilter(TriplesBlock(l ++ r), Expression(lexp ++ expr)))
	    case (Some(TriplesBlock(l)), ~(TableFilter(null, expr), Some(TriplesBlock(r)))) => Some(TableFilter(TriplesBlock(l ++ r), expr))
	    case (Some(gp             ), ~(TableFilter(null, expr), None                 )) => Some(TableFilter(gp, expr))
	    case (None,                  ~(TableFilter(null, expr), Some(TriplesBlock(r)))) => Some(TableFilter(TriplesBlock(r), expr))

	    // case (None,     ~(TableConjunction(gps), None    )) => TableConjunction(gps)
	    // case (Some(gp), ~(TableConjunction(gps), None    )) => TableConjunction(List(List(gp) ++ gps))
	    // case (None,     ~(TableConjunction(gps), Some(tb))) => TableConjunction(List(gps ++ List(tb)))
	    // case (Some(gp), ~(TableConjunction(gps), Some(tb))) => TableConjunction(List(List(gp) ++ gps ++ List(tb)))

	    case (None                    ,  ~(x, None    )) => Some(x                                     )
	    case (Some(TableConjunction(l)), ~(x, None    )) => Some(TableConjunction(l ++ List(    x    )))
	    case (Some(gp                 ), ~(x, None    )) => Some(TableConjunction(     List(gp, x    )))
	    case (None                     , ~(x, Some(tb))) => Some(TableConjunction(     List(    x, tb)))
	    case (Some(TableConjunction(l)), ~(x, Some(tb))) => Some(TableConjunction(l ++ List(    x, tb)))
	    case (Some(gp                 ), ~(x, Some(tb))) => Some(TableConjunction(     List(gp, x, tb)))

	    case x => error("found " + x)
	  }
	}).get
      }
    }
  )

  def graphpatternnottriplesORfilter:Parser[GraphPattern] = (
      "OPTIONAL"~groupgraphpattern ^^ { case "OPTIONAL"~ggp => OptionalGraphPattern(ggp) }
    | "MINUS"~groupgraphpattern ^^ { case "MINUS"~ggp => MinusGraphPattern(ggp) }
    | rep1sep(groupgraphpattern, "UNION") ^^ { x => if (x.size > 1) TableDisjunction(x) else x(0) }
    | "GRAPH"~uri~groupgraphpattern ^^ { case "GRAPH"~u~ggp => GraphGraphPattern(ggp) }
    | filter ^^ { x => TableFilter(null, x) }
  )

  def triplesblock:Parser[TriplesBlock] =
    rep1sep(triplepattern, ".") ~ opt(".") ^^ { case pats~x => TriplesBlock(pats) }

  def triplepattern:Parser[TriplePattern] =
    subject ~ predicate ~ objectt ^^ { case s~p~o => TriplePattern(s, p, o) }

  def subject:Parser[Term] = (
      qnameORuri ^^ { case x => TermUri(x) }
    | bnode ^^ { x => TermBNode(x) }
    | varr ^^ { x => TermVar(x) }
  )

  def predicate:Parser[Term] = (
      qnameORuri ^^ { x => TermUri(x) }
    | "a" ^^ { x => TermUri(Uri("http://www.w3.org/1999/02/22-rdf-syntax-ns#type")) }
    | varr ^^ { x => TermVar(x) }
  )

  def objectt:Parser[Term] = (
      qnameORuri ^^ { case x => TermUri(x) }
    | bnode ^^ { x => TermBNode(x) }
    | varr ^^ { x => TermVar(x) }
    | literal ^^ { x => TermLit(x) }
  )

  def qnameORuri:Parser[Uri] = (
      "<"~uri~">" ^^ { case "<"~x~">" => Uri(x) }
    | name~":"~name ^^ {
      case prefix~":"~localName => try {
	Uri(prefixes(prefix) + localName)
      } catch {
	case e:java.util.NoSuchElementException =>
	  throw new Exception("unknown prefix " + prefix)
      }
    }
  )

  def bnode:Parser[BNode] =
    "_:"~name ^^ { case "_:"~name => BNode(name) }

  def literal:Parser[Literal] = (
      stringLiteral~"^^"~qnameORuri ^^
      {
	case lit~"^^"~dt => Literal(rdf.TypedLiteral(lit.substring(1,lit.size - 1), dt.s match {
	  case "http://www.w3.org/2001/XMLSchema#string" => rdf.StringDatatype
	  case "http://www.w3.org/2001/XMLSchema#integer" => rdf.IntegerDatatype
	  case "http://www.w3.org/2001/XMLSchema#date" => rdf.DateDatatype
	  // case "http://www.w3.org/2001/XMLSchema#dateTime" => rdf.DateTimeDatatype
	  case x => error("only programed to deal with string and integer, not " + x)
	}))
      }
    | integer ^^ { l => Literal(rdf.TypedLiteral(l, rdf.IntegerDatatype)) }
  )

  def varr:Parser[Var] = "?"~ident ^^ { case "?"~x => Var(x) }

}

object Sparql {

}
