/* SparqlToSparql: convert SPARQL queries to sound SQL queries.
 *

 * 
 */

package org.w3.sparql2sparql

import scala.util.parsing.combinator._
import org.w3.sparql

case class MapTarget (term:sparql.Term, nm:Option[NodePattern])

case class NodePatternMap (m:Map[sparql.Var, NodePattern]) {
  def var2maptarget (t:sparql.Term) : MapTarget = {
    t match {
      case sparql.TermVar(v) => MapTarget(t, m.get(v))
      case _ => MapTarget(t, None)
    }
  }
}

case class NodePattern (iface:String, direct:String)

case class SparqlMap (construct:sparql.Construct, nodemap:NodePatternMap)

object SparqlToSparql {

  /**
   * Some tools for making pretty output, e.g. shortening strings.
   */
  var RuleLabels = scala.collection.mutable.Map[String,String]()
  var Abbreviations = scala.collection.mutable.Map[String,String]()
  def _shorten (s:String) = {
    val s2 = if (RuleLabels.contains(s)) RuleLabels(s)
	     else s
    val s3 = Abbreviations.foldLeft(s2)((s, ss) => s.replace(ss._1, ss._2))
    s3
  }


  /**
   * Substitute the terms in a triple pattern or a graph pattern with other terms.
   */
  def substituteTerm (changeMe:sparql.Term, from:sparql.Term, to:sparql.Term, nm:Option[NodePattern]):sparql.Term = {
    if (changeMe == from) {
      from match {
	case sparql.TermVar(v) => {
    	  if (nm.isDefined) {
    	    val nodeMap = nm.get
	    to match {
    	    case sparql.TermVar(v2) => {
    	      println("reverse map(" + v2 + ",\n            " + nodeMap.direct + ",\n            " + nodeMap.iface)
	      to
	    }
    	    case sparql.TermBNode(b) => {
    	      println("reverse map(" + b + ",\n            " + nodeMap.direct + ",\n            " + nodeMap.iface)
	      to
	    }
    	    case sparql.TermUri(u) => {
	      val s:String = u.s
	      val r = u.s.replaceAll(nodeMap.iface, nodeMap.direct)
	      //println("\"" + u.s + "\".replaceAll("+nodeMap.iface+", "+nodeMap.direct+") = \""+r+"\"")
	      sparql.TermUri(sparql.Uri(u.s.replaceAll(nodeMap.iface, nodeMap.direct))) // !!! failure here should error()
	    }
    	    case sparql.TermLit(l) => 
    	      error("variable " + v + " bound to literlal " + l + " but should be a node as indicated by " + nodeMap)
	    }
	  } else to
	}
	case _ => to
      }
    } else changeMe
  }

  def substitute (gp:sparql.GraphPattern, from:sparql.Term, target:MapTarget) = {
    gp match {
      case sparql.TriplesBlock(triplepatterns) => sparql.TriplesBlock(
	triplepatterns.map((tp) => {
	  // println("sub(" + tp.o + ", " + from + ", " + to + ") => ")
	  // println(substituteTerm(toTerm(tp.o), tp.o, from, target.term))
	  sparql.TriplePattern(substituteTerm(tp.s, from, target.term, target.nm),
			       tp.p,
			       substituteTerm(tp.o, from, target.term, target.nm))
	}
      ))
      case _ => error("not implemented" + gp)
    }
  }

  type VarMap = Map[sparql.Var, MapTarget]

  def substituteGraphPattern (gp:sparql.GraphPattern, vartargetmap:VarMap, varPrefix:String):sparql.GraphPattern = {
    /**
     * Substitute gp (the rule body) with the variables and constants from the
     * query which matched variables in the rule head.
     */
    val mapped = vartargetmap.foldLeft(gp)((incrementalGP, vartarget) => {
      val (varr, target) = vartarget
      val ret = substitute(incrementalGP, sparql.TermVar(varr), target)
      //println("^^incrementalGP: " + incrementalGP + "\nvartarget: " + vartarget + "\nret: " + ret)
      ret;
    })

    /**
     * "Uniquely" prefix unmapped vars to void conflict with other rules.
     */
    // val bound = Set[sparql.Var](vartargetmap.map((vartarget) => vartarget._1))
    val bound = vartargetmap.foldLeft(Set[sparql.Var]())((s, vartarget) => s + vartarget._1)
    val mappedTo = vartargetmap.foldLeft(Set[sparql.Term]())((s, vartarget) => {
      val MapTarget(term, optnp) = vartarget._2
      (term, optnp) match {
	case (sparql.TermUri(u), Some(np)) => {
	  s + sparql.TermUri(sparql.Uri(u.s.replaceAll(np.iface, np.direct))) // !!! failure here should error()
	}
	case (_, _) => s + vartarget._2.term
      }
    })
    val vars = gp.findVars
    val diff = vars -- bound
    diff.foldLeft(mapped)((incrementalGP, varr) => {
      val full = substitute(incrementalGP, sparql.TermVar(varr), 
		 MapTarget(sparql.TermVar(sparql.Var(varPrefix + varr.s)),
			   None// !! vartargetmap(varr).nm
			 ))
      full.trim(mappedTo)
    })
  }
  case class RuleIndex (trigger:sparql.TriplePattern, smap:SparqlMap) {
    override def toString = "{ \"" + trigger + "\" } => {\"\n  " + _shorten(smap.construct.gp.toString).replace("\n", "\n  ") + "\n\"}"
    def transform (tp:sparql.TriplePattern):sparql.GraphPattern = {
      substitute(substitute(smap.construct.gp, trigger.s, smap.nodemap.var2maptarget(tp.s)),
		 trigger.o, smap.nodemap.var2maptarget(tp.o))
    }
  }

  /**
   * Bindings keep track of all the ways that a particular rule has been invoked
   * in order to cover a particular query pattern. It's a mapping from a rule to
   * the list of "editions" (invocations) of that rule.
   */
  case class Bindings (b:Map[sparql.Construct, List[VarMap]]) {
    def countEditions () = {
      var count = 0
      b.map((constructlist) =>
	constructlist._2.map((_) =>
	  count = count + 1))
      count
    }
    def toGraphPattern (ident:String):sparql.GraphPattern = {
      var conjNo = 0
      val conjuncts = b.keySet.toList.flatMap(construct => {
	val l = b(construct)
	val patterns:List[sparql.GraphPattern] = l.map((vartermmap) => {
	  val unique = ident + conjNo + "_"
	  conjNo = conjNo + 1
	  substituteGraphPattern(construct.gp, vartermmap, unique)
	})
	patterns
      })

      if (conjuncts.size == 0)
	sparql.TriplesBlock(List[sparql.TriplePattern]())
      else if (conjuncts.size > 1)
	sparql.TableConjunction(conjuncts).simplify
      else
	conjuncts(0)
    }
    override def toString = "Bindings(Map(" + b.map((constructlist) => {
      val (construct, l) = constructlist
      "\"" + _shorten(construct.head.toString) + "\" -> List(" + l.map((vartermmap) => {
	"Map(" + vartermmap.map((varterm) => {
	  val (varr, term) = varterm
	  varr.toString + ":" + term.toString
	}) + ")"
      }).mkString("\n    ", ",\n    ", "") + ")"
    }).mkString("\n  ", ",\n  ", "") + "))"
    /**
     * Make sure a particular rule has an entry in the Bindings.
     */
    def ensureGraphPattern (construct:sparql.Construct) = {
      if (b.contains(construct)) this
      else Bindings(b + (construct -> List[VarMap]()))
    }
    // val varsS:Option[Bindings] = vars.maybeRebind(construct, v, tos)
    // b:Map[sparql.Construct, List[VarMap]]
    def mustBind (construct:sparql.Construct, nodemap:NodePatternMap, vs:sparql.Term, tos:sparql.Term, vo:sparql.Term, too:sparql.Term):Bindings = {
      /* ridiculous traversal for the first viably matching rule edition. */
      var matched = false
      val existing:List[VarMap] = b(construct).map((map:VarMap) => {
	def _matches (l:sparql.Term, r:sparql.Term):(Boolean, VarMap) = {
	  val empty:VarMap = Map[sparql.Var, MapTarget]()
	  (l, r) match {
	    case (v:sparql.TermVar, x) =>
	      // println("(v:sparql.TermVar, x)" + v.v + ":" + x)
	      if (map.contains(v.v)) {
		if (r == map(v.v).term) (true, empty)
		else (false, empty)
	      } else {
		(true, Map[sparql.Var, MapTarget](v.v -> MapTarget(r, nodemap.m.get(v.v))))
	      }
	    case (x, v:sparql.TermVar) => {
	      // println("query variable " + v + " known equal to " + x + " at compile time")
	      (true, empty)
	    }
	    case (x, y) =>
	      // println("(x, y)" + x + "?=" + y)
	      if (x == y) (true, empty)
	      else (false, empty)
	  }
	}
	val (sMatches, sBindings) = _matches(vs, tos)
	val (oMatches, oBindings) = _matches(vo, too)

	if (sMatches & oMatches) {
	  matched = true
	  map ++ sBindings ++ oBindings
	} else
	  map
      })
      if (matched)
	Bindings(b + (construct -> existing))
      else {
	Bindings(b.map((constructlist) => {
	  val (oldConstr, l) = constructlist
	  if (oldConstr == construct) {
	    def _newBinding (l:sparql.Term, r:sparql.Term):VarMap = {
	      val empty:VarMap = Map[sparql.Var, MapTarget]()
	      (l, r) match {
		case (v:sparql.TermVar, _) =>
		  Map[sparql.Var, MapTarget](v.v -> MapTarget(r, nodemap.m.get(v.v)))
		case (b:sparql.TermBNode, _) => {
		  println(".. synthetic query variable " + b + "")
		  Map[sparql.Var, MapTarget]()
		  // println("@@ mustBind:_newBinding(BNode) + " + b)
		  // Map(sparql.Var("bnode_" + b.b.s) -> r) // !!!
		}
		case (_, v:sparql.TermVar) => {
		  println(".. query variable " + v + " known equal to " + l + " at compile time")
		  Map[sparql.Var, MapTarget]()
		}
		case (_, _) => Map[sparql.Var, MapTarget]()
	      }
	    }
	    val ent = _newBinding(vs, tos) ++ _newBinding(vo, too)
	    val list = l ++ List(ent)
	    (construct -> list)
	  } else {
	    (oldConstr -> l)
	  }
	}))
      }
    }
  }
  def createEmptyBindings () = Bindings(Map[sparql.Construct, List[VarMap]]())

  case class RuleMap (rules:Map[sparql.Uri, List[RuleIndex]]) {
    /**
     * Recursively examine the first triple pattern in a graph pattern and add
     * invocations to the bindings to produce that triple.
     */
    def transform (prove:List[sparql.TriplePattern], used:Set[sparql.TriplePattern], varsP:Bindings):Bindings = {
      val _pad = used.foldLeft("")((s, x) => s + " ")
      def _deepPrint (s:String):Unit = {
	//print(used.size + ":" + _pad + s.replace("\n", "\n" + _pad) + "\n\n")
      }
      def _deepPrint1 (prefix:String, s:String):Unit = {
	val p = used.size + ":" + prefix + _pad
	//print(p + s.replace("\n", "\n" + p) + "\n\n")
      }

      val car = prove(0)
      val cdr = prove.filterNot (_ == car)
      _deepPrint("RuleMap.transform(" + _shorten(car.toString) + ", " + varsP.toString + ")")
      val xform = car.p match {
	case sparql.TermUri(u) =>
	  // for each rule that supplies predicate u
	  var _ruleNo = 0;
	  rules(u).foldLeft(createEmptyBindings)((bindings, hornRule) => {
	    val _prefix = "rule:" + _ruleNo
	    _ruleNo = _ruleNo + 1
	    _deepPrint1(_prefix, "trying " + _shorten(hornRule.trigger.toString))
	    val vars = varsP.ensureGraphPattern(hornRule.smap.construct)
	    // try matching the subject
	    val varss:Bindings = vars.mustBind(hornRule.smap.construct, hornRule.smap.nodemap, hornRule.trigger.s, car.s, hornRule.trigger.o, car.o)
	    val ret =
	      if (cdr.size > 0) {
		transform(cdr, used + car, varss)
	      } else {
		_deepPrint1("match!", _shorten(hornRule.trigger.toString) + "(" + _shorten(car.toString) + ") matches with " + _shorten(varss.toString))
		varss
	      }
	    _deepPrint1(_prefix, _shorten(hornRule.trigger.toString) + "(" + _shorten(car.toString) + ") matches ..." + bindings.toString + ret.toString)

	    /* Magic Huristics */
	    if (bindings.countEditions == 0) {
	      //print("choose: bindings.countEditions == 0 => " + ret + "\n\n")
	      ret
	    } else if (ret.countEditions == 0) {
	      //print("choose: ret.countEditions == 0 => " + bindings + "\n\n")
	      bindings
	    } else if (ret.countEditions < bindings.countEditions) {
	      //print("choose: "+ret.countEditions+" < "+bindings.countEditions+" => " + ret + "\n\n")
	      ret
	    } else {
	      //print("choose: "+ret.countEditions+" > "+bindings.countEditions+" => " + bindings + "\n\n")
	      bindings
	    }
	  })
	case _ => error("not implemented: " + car.p)
      }
      _deepPrint("RuleMap.transform(" + _shorten(car.toString) + ") => " + _shorten(xform.toString))
      xform
    }

    override def toString = "RuleMap(Map(" + rules.map((ent) => {
      val (u, l) = ent
      "\"" + _shorten(u.toString) + "\" -> List(" + l.map((hr) => {
	_shorten(hr.toString).replace("\n", "\n    ")
      }).mkString("\n    ", ",\n    ", "") + ")"
    }).mkString("\n  ", ",\n  ", "") + "))"

  }

  def mapGraphPattern (gp:sparql.GraphPattern, ruleMap:RuleMap, ident:String):sparql.GraphPattern = {
    gp match {
      case sparql.TriplesBlock(tps) => {
	val emptyBindings = createEmptyBindings
	val b:Bindings = ruleMap.transform(tps, Set[sparql.TriplePattern](), emptyBindings)
	//print("mapGraphPattern: " + _shorten(gp.toString) + ") => " + _shorten(b.toString) +  "\n\n")
	val g:sparql.GraphPattern = b.toGraphPattern(ident)
	//print("mapGraphPattern: ... instantiated: " + _shorten(g.toString) +  "\n\n")
	g
      }
      case sparql.TableFilter(gp2:sparql.GraphPattern, expr:sparql.Expression) =>
	sparql.TableFilter(mapGraphPattern(gp2, ruleMap, ident), expr)
      case sparql.MinusGraphPattern(gp2) =>
	sparql.MinusGraphPattern(mapGraphPattern(gp2, ruleMap, ident))
      case sparql.OptionalGraphPattern(gp2) =>
        sparql.OptionalGraphPattern(mapGraphPattern(gp2, ruleMap, ident))
      case sparql.GraphGraphPattern(gp2) =>
	sparql.GraphGraphPattern(mapGraphPattern(gp2, ruleMap, ident))
      case sparql.TableDisjunction(l) =>
	sparql.TableDisjunction({l.foldLeft((List[sparql.GraphPattern](), 0))((p, gp2) => {
	  val (l, disjNo) = p
	  (l ++ List(mapGraphPattern(gp2, ruleMap, ident + disjNo + "_")), disjNo + 1)
	})}._1)
      case sparql.TableConjunction(l) =>
	sparql.TableConjunction({l.foldLeft((List[sparql.GraphPattern](), 0))((p, gp2) => {
	  val (l, conjNo) = p
	  (l ++ List(mapGraphPattern(gp2, ruleMap, ident + conjNo + "_")), conjNo + 1)
	})}._1)
    }
  }

  def apply (query:sparql.Select, maps:List[SparqlMap]) : sparql.Select = {
    var _ruleNo = 0
    val ruleMap = RuleMap({
      maps.foldLeft(Map[sparql.Uri, List[RuleIndex]]())((m, sm) => {
	val SparqlMap(rule, nodemap) = sm
	// Register abbreviations for debugging output.
	RuleLabels.update(rule.head.toString, "head" + _ruleNo)
	RuleLabels.update(rule.gp.toString, "body" + _ruleNo)

	_ruleNo = _ruleNo + 1
	rule.head.triplepatterns.foldLeft(m)((m, tp) => m + ({
	  tp.p match {
	    case sparql.TermUri(u) => u -> {
	      if (m.contains(u)) m(u) ++ List(RuleIndex(tp, sm))
	      else List(RuleIndex(tp, sm))}
	    case _ => error("not implemented: " + tp.p)
	  }
	}))
      })
    })
    //println("ruleMap: " + ruleMap)
    sparql.Select(
      query.distinct,
      query.attrs,
      mapGraphPattern(query.gp, ruleMap, "_"),
      query.order, query.offset, query.limit
    )
  }

}

