/** SparqlToSql: convert SPARQL queries to sound SQL queries.
 * <a href="http://www.w3.org/TR/rdf-sparql-query/#sparqlDefinition">SPARQL semantics</a> defines the evalutation of an abstract query on a dataset.
 * This object maps a SPARQL abstract query over a <a href="@@">direct graph</a> to an SQL abstract query over the constituant relations.
 * <p/>
 * Please read from the bottom -- i.e. apply calls mapGraphPattern with the root
 * graph pattern. mapGraphPattern handles all the graph pattern types in SPARQL,
 * effectively peforming the Convert Graph Patterns step in SPARQL 1.0 12.2.1
 *   <a href="http://www.w3.org/TR/rdf-sparql-query/#convertGraphPattern">SPARQL rules for converting graph patterns</a>
 * with the target semantics in SQL instead of SPARQL.
 */

package org.w3.sparql2sql

import scala.util.parsing.combinator._
import java.net.URI
import org.w3.rdb
import org.w3.sql
import org.w3.sql.PrettySql.toPrettySql
import org.w3.sparql
import org.w3.sparql.rdf
import org.w3.util

case class StemURI(s:String)
case class PrimaryKey(attr:rdb.RDB.AttrName)

sealed abstract class Binding
case class RDFNode(relvarattr:sql.RelVarAttr) extends Binding
case class Str(relvarattr:sql.RelVarAttr) extends Binding
case class Int(relvarattr:sql.RelVarAttr) extends Binding
case class Enum(relvarattr:sql.RelVarAttr) extends Binding

/**
 * Converts a SPARQL object to an SQL object equivalent over the direct graph.
 *
 * @see {@link org.w3.sparql.Sparql Sparql}
 * @see {@link org.w3.sql.Sql#Select Sql}
 */
object SparqlToSql {
  /**
   * Accumulated state for generating an Sql object.
   *
   * @param joins an AddOrderedSet of SQL joins
   * @param varmap a map from Sparql.Assignable to SQL terms
   * @param exprs a set of accumulated SQL expressions
   */
  case class R2RState(joins:util.AddOrderedSet[sql.Join], varmap:Map[sparql.Assignable, SQL2RDFValueMapper], exprs:Set[sql.Expression])

  /**
   * Binding for a sparql.Variable or BNode
   */
  sealed abstract class FullOrPartialBinding
  case class FullBinding(relvarattr:sql.RelVarAttr) extends FullOrPartialBinding
  case class BindingConstraint(expr:sql.RelationalExpression, relvarattr:sql.RelVarAttr)

  /**
   * Partial binding, a variable only (so far) found in OPTIONALs or assymetric
   * UNIONs.
   */
  case class PartialBinding(binders:Set[BindingConstraint]) extends FullOrPartialBinding

  /**
   * Convert a binding to an SQL expression.
   * <p/>
   * example return:
   *   <code>if (g_union1._DISJOINT_ != 0, g_union1.who, if (g_union2._DISJOINT_ != 3, g_union2.who, NULL))</code>
   * @param against binding to be expressed
   * @return SQL expression representing that binding
   */
  def toExpr(against:FullOrPartialBinding):sql.Expression = {
    against match {
      case FullBinding(relvarattr) =>
	sql.PrimaryExpressionAttr(relvarattr)
      case PartialBinding(binders) =>
	binders.toList.reverse.foldLeft(sql.ConstNULL():sql.Expression)((exp, binding) => {
	  val BindingConstraint(expr, relvarattr) = binding
	  sql.IfElse(expr, sql.PrimaryExpressionAttr(relvarattr), exp)
	})
    }
  }
  /**
   * Accumulate bindings on previously bound sparql variables.
   * @param binding previous binding
   * @param relVarAttr SQL relvar attribute to be bound, e.g. <code>G_union1.who</code>
   * @param expr expr binding constraint, e.g. <code>G_opt6._DISJOINT_ IS NULL</code> or <code>G_union1._DISJOINT_!=0</code>
   * @return PartialBinding on old constraints plus expr=relVarAttr
   */
  def addExpr(binding:FullOrPartialBinding, relVarAttr:sql.RelVarAttr, expr:sql.RelationalExpression):FullOrPartialBinding = {
    binding match {
      case FullBinding(relvarattr) =>
	error("Unexpected FullBinding to " + relvarattr + "\n" + 
	      "because subselectVars should only find a pre-existing PartialBindings.")
        // return binding if this codepath is needed later.
      case PartialBinding(binders) =>
	/**
	 * Add a new BindingConstraint to existing one.
	 * e.g. existing constraint: G_union1._DISJOINT_!=0,G_union1.name
	 * add for second side of union: G_union1._DISJOINT_!=1,G_union1.name
	 */
	PartialBinding(binders + BindingConstraint(expr, relVarAttr))
    }
  }

  /**
   * SQL terms representing SPARQL variables and bnodes.
   */
  sealed abstract class SQL2RDFValueMapper(binding:FullOrPartialBinding)
  case class IntMapper(binding:FullOrPartialBinding) extends SQL2RDFValueMapper(binding)
  case class StringMapper(binding:FullOrPartialBinding) extends SQL2RDFValueMapper(binding)
  case class DateMapper(binding:FullOrPartialBinding) extends SQL2RDFValueMapper(binding)
  /**
   * map to a URL for a tuple in the database.
   */
  case class RDFNoder(relation:rdb.RDB.RelName, binding:FullOrPartialBinding) extends SQL2RDFValueMapper(binding)
  /**
   * map to a blank node label for a tuple in the database.
   */
  case class RDFBNoder(relation:rdb.RDB.RelName, binding:FullOrPartialBinding) extends SQL2RDFValueMapper(binding)

  /**
   * a URL representing a tuple in a database.
   */
  case class NodeUri(stem:Stem, rel:Rel, attr:Attr, v:CellValue)

  /**
   * url stem to base the direct graph.
   * <p/>
   * e.g. http://mydb.example/theDatabase
   */
  case class Stem(s:String) {
    override def toString = "" + s
  }

  /**
   * SQL relation (table) name
   */
  case class Rel(s:String) {
    override def toString = "" + s
  }
  /**
   * SQL attribute (column) name
   */
  case class Attr(s:String) {
    override def toString = "" + s
  }
  /**
   * value in a database
   */
  case class CellValue(s:String)
  /**
   * a URL representing a predicate in a database.
   */
  case class PUri(stem:Stem, rel:Rel, attr:Attr) {
    override def toString = "<" + stem + "/" + rel + "#" + attr + ">"
  }
  /**
   * parse predicate URL in direct graph into stem, relation and attribute
   * <pre>stemURI + '/' + (\w+) + '#' (\w+)</pre>
   */
  def parsePredicateURI(u:sparql.Uri):PUri = {
    val x:String = u.s
    val uri = new URI(x)
    val path = uri.getPath().split("/").toList.filterNot(_ == "")
    val subPath = path.slice(0, path.size - 1).mkString("/")
    val stem = uri.getScheme() + "://" + uri.getAuthority + "/" + subPath
    PUri(Stem(stem), Rel(path.last), Attr(uri.getFragment))
  }

  /**
   * parse node URL in direct graph into stem, relation, attribute and value
   * <pre>stemURI + '/' (\w+) '/' (\w+) '.' (\w+) '#record'</pre>
   */
  def parseObjectURI(u:sparql.Uri):NodeUri = {
    try {
      val x:String = u.s
      val uri = new URI(x)
      val path = uri.getPath().split("/").toList.filterNot(_ == "")
      val subPath = path.slice(0, path.size - 2).mkString("/")
      val rel = path(path.size - 2)
      val attrPair = path(path.size-1).split("\\.")
      val stem = uri.getScheme() + "://" + uri.getAuthority + "/" + subPath
      assert("record" == uri.getFragment)
      NodeUri(Stem(stem), Rel(rel), Attr(attrPair(0)), CellValue(attrPair(1)))
    } catch {
      case e:java.lang.IndexOutOfBoundsException =>
	throw new Exception("unable to parse " + u + " as an object")
    }
  }
  /**
   * synthesize a relvar name from a SPARQL term.
   * <p/>
   * e.g. <code>?emp</code> =&gt;<code>R_emp</code>
   * e.g. <code>&lt;http://hr.example/DB/Employee/empid.253#record&gt;</code> =&gt;<code>R_empid253</code>
   * e.g. <code>18</code> =&gt;<code>R_18</code>
   */
  def relVarFromTerm(s:sparql.Term):sql.RelVar = {
    s match {
      case sparql.TermUri(ob) => relVarFromNode(ob)
      case sparql.TermVar(v)  => relVarFromVar(v)
      case sparql.TermBNode(v)  => relVarFromBNode(v)
      case sparql.TermLit(l)  => relVarFromLiteral(l)
    }
  }

  def relVarFromNode(u:sparql.Uri):sql.RelVar = {
    val NodeUri(stem, rel, Attr(a), CellValue(v)) = parseObjectURI(u)
    sql.RelVar(sql.Name("R_" + a + v))
  }

  def relVarFromLiteral(l:sparql.Literal):sql.RelVar = {
    sql.RelVar(sql.Name("R_" + l.lit.lexicalForm))
  }

  def relVarFromVar(vr:sparql.Var):sql.RelVar = {
    val sparql.Var(v) = vr
    sql.RelVar(sql.Name("R_" + v))
  }

  def relVarFromBNode(vr:sparql.BNode):sql.RelVar = {
    val sparql.BNode(b) = vr
    sql.RelVar(sql.Name("B_" + b))
  }

  /**
   * synthesize a SQL name from a SPARQL variable or bnode.
   * <p/>
   * e.g. <code>?emp</code> =&gt;<code>emp</code>
   */
  def attrAliasNameFromVar(v:sparql.Assignable):String = "" + v.s

  /**
   * add constraints implied by a URI
   * @param state state to be appended
   * @param constrainMe relvar attribute to constrain, e.g. <code>R_empid18.empid</code>
   * @param u SparqlToSql URL object, e.g. <code>NodeUri(http://hr.example/DB,Employee,empid,CellValue(18))</code>
   * @return state + expression for the URI
   * <p/>
   * <code>http://hr.example/DB/Employee/empid=18, true</code> =&gt;<code>R_emp.manager=18</code>
   * <code>http://hr.example/DB/Employee/empid=18, false</code> =&gt;<code>R_empid18.empid=18</code>
   * (the latter produces another join on R_empid18 in order to enforce a foreign key.)
   * http://hr.example/DB/Employee/empid=18 has already been parsed to produce a NodeUri:
   *   NodeUri(http://hr.example/DB,Employee,empid,CellValue(18))
   */
  def uriConstraint(state:R2RState, constrainMe:sql.RelVarAttr, u:NodeUri):R2RState = {
    R2RState(state.joins,
	     state.varmap,
	     state.exprs + sql.RelationalExpressionEq(sql.PrimaryExpressionAttr(constrainMe),
						      sql.PrimaryExpressionTyped(rdb.RDB.Datatype.INTEGER,sql.Name(u.v.s))))
  }

  /**
   * add constraints implied by a literal in a SPARQL triple pattern
   * @param state state to be appended
   * @param constrainMe relvar attribute to constrain, e.g. <code>R_18.empid</code>
   * @param lit sparq.Literal, e.g. <code>18</code>
   * @param dt sparq datatype, e.g. <code>xsd:integer</code>
   * @return state + expression for the URI, e.g. <code>R_18.empid=18</code>
   */
  def literalConstraint(state:R2RState, constrainMe:sql.RelVarAttr, lit:sparql.Literal, dt:rdb.RDB.Datatype):R2RState = {
    R2RState(state.joins,
	     state.varmap,
	     state.exprs + sql.RelationalExpressionEq(sql.PrimaryExpressionAttr(constrainMe),
						      sql.PrimaryExpressionTyped(dt,sql.Name(lit.lit.lexicalForm))))
  }

  /** varConstraint
   * @param state  earlier R2RState
   * @param alias  relvar to bind, e.g. Employees AS R_emp
   * @param optAttr  SQL attribute to bind, e.g. Employees.lastName
   * @param v  SPARQL variable or blank node to bind
   * @param db  database description
   * @param rel  SQL relation to bind, e.g. Employee
   * @return a new R2RState incorporating the new binding
   * For example, <code>SELECT ?emp WHERE { ?emp emp:lastName ?name }</code> will call varConstraint twice:
   * 
   *    given: alias:=R_emp, optAttr:=lastName, v:=?name, rel:=Employee -&gt;
   *   return: (VarAssignable(?name),StringMapper(FullBinding(R_emp.lastName)))
   *   which maps "Smith" to "Smith"^^xsd:string
   * 
   *    given: alias:=R_emp, optAttr:=empid, v:=?emp, rel:=Employee -&gt;
   *   return: (VarAssignable(?emp),RDFNoder(Employee,FullBinding(R_emp.empid)))
   *   which maps 4 to &lt;http://hr.example/our/favorite/DB/Employee/id.4#record&gt;
   */
  def varConstraint(state:R2RState, alias:sql.RelVar, optAttr:Option[rdb.RDB.AttrName], v:sparql.Assignable, db:rdb.RDB.Database, rel:rdb.RDB.RelName):R2RState = {
    val constrainMe = if (optAttr.isDefined) sql.RelVarAttr(alias, optAttr.get) else sql.RelVarAttr(alias, rdb.RDB.AttrName("_no_such_attribute"))
    val reldesc = db(rel)
    val boundTo = FullBinding(constrainMe)

    /**
     * Bind optAttr to an SQL generator like RDFNoder(Employee,FullBinding(R_emp.empid))
     */
    val binding = reldesc.pk match {
      /** <pre>varConstraint(R_emp, Some(empid), VarAssignable(?emp), Employee) -&gt; RDFNoder(Employee,FullBinding(R_emp.empid))</pre> */
      case Some(rdb.RDB.CandidateKey(List(rdb.RDB.AttrName(constrainMe.attribute.n)))) => RDFNoder(rel, boundTo)
      case _ => {
	val asKey = List(constrainMe.attribute)
	if (reldesc.fks contains rdb.RDB.CandidateKey(asKey)) { // !! (0)
	  /** varConstraint(R_patient, Some(SexDE), VarAssignable(?_0_sexEntry), Person) -&gt; RDFNoder(Person,FullBinding(R_patient.SexDE)) */
          // WTF? I don't see where the following is used
	  // val rdb.RDB.Target(fkrel, fkattr) = reldesc.fks(rdb.RDB.ForeignKeyKey(asKey))
	  RDFNoder(rel, boundTo)
	} else if (reldesc.header isDefinedAt constrainMe.attribute) {
	  reldesc.header(constrainMe.attribute) match {
	    /** varConstraint(R__0_indicDE, Some(NDC), VarAssignable(?_0_indicNDC), Medication_DE) -&gt; IntMapper(FullBinding(R__0_indicDE.NDC)) */
	    case rdb.RDB.Datatype("Int") => IntMapper(boundTo)
	    /** varConstraint(R_emp, Some(lastName), VarAssignable(?name), Employee) -&gt; StringMapper(FullBinding(R_emp.lastName)) */
	    case rdb.RDB.Datatype("String") => StringMapper(boundTo)
	    /** varConstraint(R_patient, Some(DateOfBirth), VarAssignable(?dob), Person) -&gt; DateMapper(FullBinding(R_patient.DateOfBirth)) */
	    case rdb.RDB.Datatype("Date") => DateMapper(boundTo)
	  }
	} else {
	  /** Default behavior for unknown attributes. */
	  RDFBNoder(rel, boundTo) // @@ untested
	} 
      }
    }

    if (state.varmap.contains(v) && state.varmap(v) == constrainMe) {
      /**
       * No useful contributions for this variable.
       * (We could re-add the binding; this case is just for clarity of behavior.)
       */
      state
    } else if (state.varmap.contains(v)) {
      /**
       * The variable has already been bound to another attribute.
       * Constraint against the initial binding for this variable.
       * e.g. <code>R__0_0_indicCode.ID=R__0_0_indicCode.ID</code>
       */
      val constraint = sql.RelationalExpressionEq(sql.PrimaryExpressionAttr(constrainMe),
						  sql.PrimaryExpressionAttr(varToAttribute(state.varmap, v)))
      R2RState(state.joins, state.varmap, 
	       if (varToAttributeDisjoints(state.varmap, v).size > 0)
		 /**
		  * Enumerate the set of constraints capturing all sides of a disjoint or option.
		  * e.g. a UNION where two disjoints constrain R_who.empid=G_union0.who:
		  * Set((G_union0._DISJOINT_!=0) OR (R_who.empid=G_union0.who),
		  *     (G_union0._DISJOINT_!=1) OR (R_who.empid=G_union0.who))
		  */
		 state.exprs ++ {varToAttributeDisjoints(state.varmap, v) map ((d) => sql.ExprDisjunction(Set(d, constraint)))}
	       else
		 state.exprs + constraint
	     )
    } else {
      /**
       * Add binding for new variable.
       */
      R2RState(state.joins, state.varmap + (v -> binding), state.exprs)
    }
  }

  def toString(relvarattr:sql.RelVarAttr) : String = {
    relvarattr.relvar.n.s + "." + relvarattr.attribute.n
  }
  // def toString(mapper:SQL2RDFValueMapper) : String = {
  //   mapper match {
  //     case IntMapper(relvar, disjoints) => "INT: " + toString(relvar)
  //     case StringMapper(relvar, disjoints) => "STRING: " + toString(relvar)
  //     case DateMapper(relvar, disjoints) => "DATE: " + toString(relvar)
  //     case RDFNoder(relation, relvar, disjoints) => "RDFNoder: " + relation.n.s + ", " + toString(relvar)
  //     case RDFBNoder(relation, relvar, disjoints) => "RDFBNoder: " + relation.n.s + ", " + toString(relvar)
  //   }
  // }

  def __optFirst (k:Option[rdb.RDB.CandidateKey]):Option[rdb.RDB.AttrName] =
    k match {
      case Some(ck) => Some(ck.attrs(0))
      case _ => None
    }

  /**
   * map a given triple to one or two joined tables, variable
   * @param db  database description
   * @param stateP  earlier R2RState
   * @param triple  
   * @param enforceForeignKeys  
   * @return a new R2RState incorporating the new binding
   * bindings to RelVarAttrs, and constraints if those variables were
   * already bound. */
  def bindOnPredicate(db:rdb.RDB.Database, stateP:R2RState, triple:sparql.TriplePattern, enforceForeignKeys:Boolean):R2RState = {
    val sparql.TriplePattern(s, p, o) = triple
    p match {
      /** Don't deal with ?s ?p ?o . We could deal with <s> ?p ?o , but haven't yet. */
      case sparql.TermVar(v) => error("variable predicates require tedious enumeration; too tedious for me.")
      case sparql.TermUri(uri) => {
	val PUri(stem, spRel, spAttr) = parsePredicateURI(uri)
	/** The relation and attribute come from the predicate, e.g. Employee.fname =&gt; Employee and fname. */
	val rel = rdb.RDB.RelName(spRel.s)
	val attr = rdb.RDB.AttrName(spAttr.s)

	/**
	 * The particular join for e.g. Employee is controled by the subject.
	 * ?emp =&gt; Employee AS emp
	 * <Employee:18> =&gt; Employee AS Employee_18 .
         */
	val relvar = relVarFromTerm(s)
	/** Synthesize relvar attribute, e.g. emp.fname. */
	val objattr = sql.RelVarAttr(relvar, attr)

	/** Accumulate joins and constraints that come from the subject */
	val state_postSubj = s match {
	  case sparql.TermUri(u) =>
	    /** additional constraint, e.g. R_empid18.empid=18. */
	    uriConstraint(stateP, sql.RelVarAttr(relvar, db(rel).pk.get.attrs(0)), parseObjectURI(u)) // !! (0)
	  case sparql.TermVar(v) =>
	    /** assignable binding for novel vars, new constraint for previously bound vars. */
	    try {
	      varConstraint(stateP, relvar, __optFirst(db(rel).pk), sparql.VarAssignable(v), db, rel) // !! (0)
	    } catch {
	      case e:java.util.NoSuchElementException =>
		/** Tell user that the relation.attribute encoded in the subject was not found in the database description. */
		throw new Exception("error processing { " + s + " " + p + " " + o + " } :db(" + rel + ") not found in " + db)
	    }
	  case sparql.TermBNode(b) =>
	    /** assignable binding for novel bnodes, new constraint for previously bound bnodes. */
	    try {
	      varConstraint(stateP, relvar, __optFirst(db(rel).pk), sparql.BNodeAssignable(b), db, rel) // !! (0)
	    } catch {
	      case e:java.util.NoSuchElementException =>
		throw new Exception("error processing { " + s + " " + p + " " + o + " } :db(" + rel + ") not found in " + db)
	    }
	  case _                 => error("illegal SPARQL subject: " + s)
	}

	/** Join rel (relation dictated by predicate) AS relvar (alias dicated by subject).
	 * may be redundant as R2RState's joins are a set */
	val state_subjJoin = R2RState(state_postSubj.joins + sql.InnerJoin(sql.AliasedResource(rel,relvar), None), state_postSubj.varmap, state_postSubj.exprs)

	try { db(rel).header(attr) } catch {
	  /** Tell user that the queried attribute was not found in the database description. */
	  case e:java.util.NoSuchElementException =>
	    throw new Exception("error processing { " + s + " " + p + " " + o + " } :db(" + rel + ").header(" + attr + ") not found in " + db)
	}

	/**
	 * fkrel.fkattr (e.g. Employee.manager) may be a foreign key.
	 * Calculate final relvarattr and relation.
	 */
	val asKey = rdb.RDB.AttrList(attr)
	val (targetattr:sql.RelVarAttr, targetrel, dt, state_fkeys:R2RState) =
	  if (db(rel).fks contains asKey) { // !! (0)
	    val rdb.RDB.Target(targetRel, targetCK) = db(rel).fks targetOf asKey
	    try { db(targetRel).header isDefinedAt targetCK(0) } catch { // !! (0)
	      /** Foreign key relation.attribute was not found in the database description. */
	      case e:java.util.NoSuchElementException =>
		throw new Exception("db(" + targetRel + ").header(" + targetCK + ") not found in " + db)
	    }
	    val fkdt =
	      if (db(targetRel).fks contains targetCK) {
	      /** Foreign key to something which is a foreign key. May have use
	       * cases, but signal error until we figure out that they are. */
		val rdb.RDB.Target(dfkrel, dfkattr) = db(targetRel).fks targetOf targetCK
		error("foreign key " + rel.n + "." + attr.n + 
		      "->" + targetRel.n + "." + targetCK(0).n +  // !! (0)
		      "->" + dfkrel.n + "." + dfkattr(0).n) // !! (0)
	      } else
		db(targetRel).header(targetCK(0)) // !! (0)
	    if (enforceForeignKeys) {
	      /**
	       * Create an extra join on the foreign key relvar. For instance,
	       * <code>?task1 <http://hr.example/DB/TaskAssignments#employee> ?who</code>
	       * where TaskAssignments.employee is a foreign key to Employee.empid
	       * will join Employee AS R_who, constrain R_who.empid=R_task1.employee
	       * and bind targetattr:R_who.empid. targetrel:Employee .
	       */
	      val oRelVar = relVarFromTerm(o)
	      val fkaliasattr = sql.RelVarAttr(oRelVar, targetCK(0)) // !! (0)
	      val state_t = R2RState(state_subjJoin.joins + sql.InnerJoin(sql.AliasedResource(targetRel,oRelVar), None),
				     state_subjJoin.varmap,
				     state_subjJoin.exprs + sql.RelationalExpressionEq(sql.PrimaryExpressionAttr(fkaliasattr),
										       sql.PrimaryExpressionAttr(objattr)))
	      //println("enforceFKs: <code>"+s+" "+p+" "+o+"</code> where "+rel+"."+attr+" is a foreign key to "+fkrel+"."+fkattr+" will join "+fkrel+" AS "+oRelVar+", constrain "+fkaliasattr+"="+objattr+" and bind targetattr:=" + fkaliasattr + ". targetrel:=" + fkrel + " (instead of " + objattr + ", " + rel + ").")
	      (fkaliasattr, targetRel, fkdt, state_t)
	    } else {
	      /**
	       * We're not enforcing foreign keys, so just bind 
	       * targetattr:=R_task1.employee, targetrel:=TaskAssignments.
	       */
	      (objattr, rel, fkdt, state_subjJoin)
	    }
	  }
	else
	  /** not a foreign key, so use the relvarattr and relation calculated
	   * from the predicate. */
	  (objattr, rel, db(rel).header(attr), state_subjJoin)

	o match {
	  case sparql.TermLit(l) => literalConstraint(state_fkeys, targetattr, l, dt)
	  case sparql.TermUri(u) => uriConstraint    (state_fkeys, targetattr, parseObjectURI(u))
	  case sparql.TermVar(v) => varConstraint    (state_fkeys, targetattr.relvar, Some(targetattr.attribute), sparql.VarAssignable(v), db, targetrel)
	  case sparql.TermBNode(b) => varConstraint  (state_fkeys, targetattr.relvar, Some(targetattr.attribute), sparql.BNodeAssignable(b), db, targetrel)
	}
      }
      case _                  => error("illegal SPARQL predicate: " + p)
    }
  }

  def bindingConstraintToAttribute(constraint:BindingConstraint):sql.RelVarAttr = {
    val BindingConstraint(expr:sql.RelationalExpression, relvarattr:sql.RelVarAttr) = constraint;
    relvarattr
  }
  def bindingToAttribute(binding:FullOrPartialBinding):sql.RelVarAttr = {
    binding match {
      case FullBinding(relvarattr:sql.RelVarAttr) => relvarattr
      case PartialBinding(binders) => bindingConstraintToAttribute(binders.toList(0))
    }
  }
  def varToAttribute(varmap:Map[sparql.Assignable, SQL2RDFValueMapper], vvar:sparql.Assignable):sql.RelVarAttr = {
    val mapper = try { varmap(vvar) } catch {
      case e:java.util.NoSuchElementException =>
	throw new Exception("mapper for variable " + vvar + " not found in " + varmap)
    }
    mapper match {
      case IntMapper(binding) => bindingToAttribute(binding)
      case StringMapper(binding) => bindingToAttribute(binding)
      case DateMapper(binding) => bindingToAttribute(binding)
      case RDFNoder(relation, binding) => bindingToAttribute(binding)
      case RDFBNoder(relation, binding) =>  bindingToAttribute(binding) // error("BNode should not need relvar " + relvar)
    }
  }

  def bindingConstraintToExpression(constraint:BindingConstraint):sql.RelationalExpression = {
    val BindingConstraint(expr:sql.RelationalExpression, relvarattr:sql.RelVarAttr) = constraint;
    expr
  }
  def bindingToDisjoints(binding:FullOrPartialBinding):Set[sql.RelationalExpression] = {
    binding match {
      case FullBinding(relvarattr:sql.RelVarAttr) => Set[sql.RelationalExpression]()
      case PartialBinding(binders) => binders.map({b => bindingConstraintToExpression(b)})
    }
  }
  def varToAttributeDisjoints(varmap:Map[sparql.Assignable, SQL2RDFValueMapper], vvar:sparql.Assignable):Set[sql.RelationalExpression] = {
    varmap(vvar) match {
      case IntMapper(binding) => bindingToDisjoints(binding)
      case StringMapper(binding) => bindingToDisjoints(binding)
      case DateMapper(binding) => bindingToDisjoints(binding)
      case RDFNoder(relation, binding) => bindingToDisjoints(binding)
      case RDFBNoder(relation, binding) =>  bindingToDisjoints(binding) // error("BNode should not need relvar " + relvar)
    }
  }

  /**
   * Converts a variable bound to a URL to an SQL expression for that URL.
   *
   * @param  varmap  map from variable to SQL2RDFValueMapper
   * @param  vvar    the variable to be represented
   * @return         an SQL CONCAT expression
   * @see            VarMap
   */
  def varToExpr(varmap:Map[sparql.Assignable, SQL2RDFValueMapper], vvar:sparql.Assignable, stem:StemURI):sql.Expression = {
    varmap(vvar) match {
      case IntMapper(binding) => sql.PrimaryExpressionAttr(bindingToAttribute(binding))
      case StringMapper(binding) => 
	sql.Concat(List(sql.PrimaryExpressionTyped(rdb.RDB.Datatype("String"),sql.Name("'")),
		    sql.PrimaryExpressionAttr(bindingToAttribute(binding)),
		    sql.PrimaryExpressionTyped(rdb.RDB.Datatype("String"),sql.Name("'^^<http://www.w3.org/2001/XMLSchema#string>"))))
      case DateMapper(binding) => sql.PrimaryExpressionAttr(bindingToAttribute(binding))
      case RDFNoder(relation, binding) => 
	sql.Concat(List(sql.PrimaryExpressionTyped(rdb.RDB.Datatype("String"),sql.Name(stem.s)),
		    sql.PrimaryExpressionTyped(rdb.RDB.Datatype("String"),relation.n),
		    sql.PrimaryExpressionTyped(rdb.RDB.Datatype("String"),sql.Name("/")),
		    sql.PrimaryExpressionTyped(rdb.RDB.Datatype("String"),bindingToAttribute(binding).attribute.n),
		    sql.PrimaryExpressionTyped(rdb.RDB.Datatype("String"),sql.Name(".")),
		    sql.PrimaryExpressionAttr(bindingToAttribute(binding)),
		    sql.PrimaryExpressionTyped(rdb.RDB.Datatype("String"),sql.Name("#record"))))
      case RDFBNoder(relation, binding) => 
	sql.Concat(List(sql.PrimaryExpressionTyped(rdb.RDB.Datatype("String"),sql.Name("_:")),
		    sql.PrimaryExpressionTyped(rdb.RDB.Datatype("String"),relation.n),
		    sql.PrimaryExpressionTyped(rdb.RDB.Datatype("String"),sql.Name(".")),
		    sql.PrimaryExpressionTyped(rdb.RDB.Datatype("String"),bindingToAttribute(binding).attribute.n),
		    sql.PrimaryExpressionTyped(rdb.RDB.Datatype("String"),sql.Name(".")),
		    sql.PrimaryExpressionAttr(bindingToAttribute(binding))))
    }
    
  }

  /**
   * return an SQL PrimaryExpression for rTerm.
   *
   * @param  varmap  map from variable to SQL2RDFValueMapper
   * @param  l  an SQL relvarattr for a SPARQL Assignable (variable or bnode)
   * @param  rTerm  SPARQL term to be converted to a PrimaryExpression
   * @param  sqlexpr  an unbound SQL relational expression, e.g. sql.RelationalExpressionEq(_,_)
   * @return   an SQL expression on the relvarattrs to which the variables in f are bound
   */
  def assignable2expr(varmap:Map[sparql.Assignable, SQL2RDFValueMapper], l:sql.RelVarAttr, rTerm:sparql.Term, sqlexpr:(sql.PrimaryExpression, sql.PrimaryExpression) => sql.RelationalExpression):sql.RelationalExpression = { // :sparql.Var
    val r:sql.PrimaryExpression = rTerm match {
      case sparql.TermUri(u) => error("not implemented: translating RDF URI to SQL: " + u) // :sparql.Uri
      case sparql.TermVar(v) => sql.PrimaryExpressionAttr(varToAttribute(varmap, sparql.VarAssignable(v)))
      case sparql.TermBNode(b) => sql.PrimaryExpressionAttr(varToAttribute(varmap, sparql.BNodeAssignable(b)))
      case sparql.TermLit(sparql.Literal(p:rdf.PlainLiteral)) => error("not implemented")
      case sparql.TermLit(sparql.Literal(rdf.TypedLiteral(lit,dt))) =>
	sql.PrimaryExpressionTyped({
	  dt match {
	    case rdf.IRI("http://www.w3.org/2001/XMLSchema#string") => rdb.RDB.Datatype.STRING
	    case rdf.IRI("http://www.w3.org/2001/XMLSchema#integer") => rdb.RDB.Datatype.INTEGER
	    case rdf.IRI("http://www.w3.org/2001/XMLSchema#date") => rdb.RDB.Datatype.DATE
	    case _ => error("unable to translate to RDF literal SQL: \"" + lit + "\"^^<" + dt + ">")
	  }
	}, lit)
    }
    sqlexpr(sql.PrimaryExpressionAttr(l), r)
  }

  /**
   * create an SQL expression analogous to a SPARQL expression. The variables in
   * f are expressed as the relvarattrs to which they are mapped. This function
   * is only minimally implemented here.
   *
   * @param  varmap  map from variable to SQL2RDFValueMapper
   * @param  f  SPARQL Expression
   * @return   an SQL expression on the relvarattrs to which the variables in f are bound
   */
  def filter2expr(varmap:Map[sparql.Assignable, SQL2RDFValueMapper], f:sparql.PrimaryExpression):sql.RelationalExpression = {

    /** sqlexpr: an unbound RelationalExpression, i.e. a function which takes
     * (sql.RelVarAttr, sql.PrimaryExpressionAttr) and returns an
     * sql.RelationalExpression
     */
    val (lTerm:sparql.Term, rTerm:sparql.Term, sqlexpr) = f match {
      case sparql.PrimaryExpressionEq(l, r) => (l.term, r.term, sql.RelationalExpressionEq(_,_))
      case sparql.PrimaryExpressionLt(l, r) => (l.term, r.term, sql.RelationalExpressionLt(_,_))
      case sparql.PrimaryExpressionGt(l, r) => (l.term, r.term, sql.RelationalExpressionGt(_,_))
      case sparql.SparqlTermExpression(t) => error("not implemented") // (l.term, r.term, sql.PrimaryExpressionAttr(_,_))
    }

    lTerm match {
      /** does not handle FILTER (<x> = ?v) */
      case sparql.TermUri(obj) => error("only SPARQL PrimaryExpressions with a variable on the left have been implemented: punting on " + f)
      /** FILTER (?v = <x> && ?v = ?x && ?v = 7) */
      case sparql.TermVar(v) => assignable2expr(varmap, varToAttribute(varmap, sparql.VarAssignable(v)), rTerm, sqlexpr)
      case sparql.TermBNode(b) => assignable2expr(varmap, varToAttribute(varmap, sparql.BNodeAssignable(b)), rTerm, sqlexpr)
      /** does not handle FILTER (7 = ?v) */
      case sparql.TermLit(lit) => error("only SPARQL PrimaryExpressions with a variable on the left have been implemented: punting on " + f)
    }
  }

  /**
   * Promote a variable in an OPTIONAL or UNION subselect to the outer
   * varmap/expressions. 
   */
  def subselectVars(startState:R2RState, v:sparql.Assignable, optionalAlias:sql.RelVar,
		    optionalCond:sql.RelationalExpression,
		    outerVarmap:Map[sparql.Assignable, SQL2RDFValueMapper],
		    nestedVarmap:Map[sparql.Assignable, SQL2RDFValueMapper],
		    isOpt:Boolean):R2RState = {

    val varAliasAttr = sql.RelVarAttr(optionalAlias, rdb.RDB.AttrName(attrAliasNameFromVar(v)))
    if (startState.varmap.contains(v)) {

      /** The variable has already been bound. */
      val newMap:Map[sparql.Assignable, SQL2RDFValueMapper] =
	/** If var was already bound to the same relvarattr... */
	if (varToAttribute(startState.varmap, v) == varAliasAttr) {
	  /**
	   * Add new partial constraint to old partial constraints to produce a new binding.
	   * example:
	   *   ?name -> StringMapper(PartialBinding(Set(BindingConstraint(G_union1._DISJOINT_!=0,G_union1.name),
           *                                            BindingConstraint(G_union1._DISJOINT_!=1,G_union1.name))))
	   */
	  Map(v -> {
	    startState.varmap(v) match {
	      case IntMapper(binding)      => IntMapper(addExpr(binding, varAliasAttr, optionalCond))
	      case StringMapper(binding)   => StringMapper(addExpr(binding, varAliasAttr, optionalCond))
	      case DateMapper(binding)     => DateMapper(addExpr(binding, varAliasAttr, optionalCond))
	      case RDFNoder(rel, binding)  => RDFNoder(rel, addExpr(binding, varAliasAttr, optionalCond))
	      case RDFBNoder(rel, binding) => RDFBNoder(rel, addExpr(binding, varAliasAttr, optionalCond))
	    } } )
	} else {
	  /** Var was bound to a different relvarattr so handle as newConstraint below. */
	  Map()
	}

      val newConstraints =
	if (varToAttribute(outerVarmap, v) == varAliasAttr) { // also varToAttribute(startState.varmap, v) == varAliasAttr
	  Set()
	} else {
	  /** Constraint against binding from earlier graph pattern.
	   * e.g. G_union1.who=R_who.empid */
	  val constraint = sql.RelationalExpressionEq(sql.PrimaryExpressionAttr(varAliasAttr),
						      sql.PrimaryExpressionAttr(varToAttribute(outerVarmap, v)))
	  if (varToAttributeDisjoints(outerVarmap, v).size > 0) {
	    /**
	     * Construct a conjunction of disjunctions like:
	     * (union0._DISJOINT_ != 0 OR union0.x=union1.x) AND (union1._DISJOINT_ != 2 OR union0.x=union1.x)
	     */
	    varToAttributeDisjoints(outerVarmap, v) map ((d) =>
	      sql.ExprDisjunction({
		if (isOpt) Set(d, constraint) // a disjunction like: G_opt1._DISJOINT_ IS NULL OR G_opt3.birthday=G_opt1.birthday
		else Set(sql.ExprConjunction(Set(d, optionalCond)), constraint) // @@ untested code path
	      }))
	  } else {
	    if (isOpt)
	      /** just the constraint from above */
	      Set(constraint)
	    else
	      /**
	       * Above constraint applied only for this path:
	       * for example: (G_union1._DISJOINT_!=0) OR (G_union1.who=R_who.empid)
	       */
	      Set(sql.ExprDisjunction(Set(optionalCond, constraint)))
	  }
	}

      R2RState(startState.joins, startState.varmap ++ newMap, startState.exprs ++ newConstraints)
    } else {
      /**
       * This variable is new to the outer context so synthesize a new partial
       * binding à la:
       *   ?name ->
       *     StringMapper(PartialBinding((G_union1._DISJOINT_!=0,G_union1.name)))
       */
      val p = PartialBinding(Set(BindingConstraint(optionalCond, varAliasAttr)))
      val mapper:SQL2RDFValueMapper = nestedVarmap(v) match {
	case IntMapper(_)      => IntMapper(p)
	case StringMapper(_)   => StringMapper(p)
	case DateMapper(_)   => DateMapper(p)
	case RDFNoder(rel, _)  => RDFNoder(rel, p)
	case RDFBNoder(rel, _) => RDFBNoder(rel, p)
      }

      R2RState(startState.joins, startState.varmap + (v -> mapper), startState.exprs)
    }
  }

  /**
   * Recursively add the joins, variable mappings and constraints for an SQL query implementing a graph pattern.
   * @param db  database description.
   * @return a new state including the subquery representing gp in a join.
   * This is factored out of mapGraphPattern as it is used for MINUS and probably later for NOT EXISTS.
   */
  def synthesizeOuterJoin(initState:R2RState, gp:sparql.GraphPattern, negate:Boolean, db:rdb.RDB.Database, enforceForeignKeys:Boolean):R2RState = {
    /** SPARQL OPTIONALs and UNIONs are treated as SQL subselects.
     * Set up initial state for this subselect.
     */
    val leftJoinAlias = sql.RelVar(sql.Name("G_opt" + initState.joins.size))
    val initDisjoints:Set[sql.Select] = Set()
    val emptyState = R2RState(
      util.AddOrderedSet[sql.Join](), 
      Map[sparql.Assignable, SQL2RDFValueMapper](), 
      Set[sql.Expression]()
    )

    /** Create the select for the nested graph pattern.
     */
    val nestedState = mapGraphPattern(db, emptyState, gp, enforceForeignKeys)
    val nestedVars = gp.findVars
    /**
     * Select a constant as _DISJOINT_ so later constraints can be
     * sensitive to whether a variable was bound.
     * This matters for assymetric UNIONs and, here, OPTIONALs. Given:
     *   Join( LeftJoin( BGP(),
     *                   BGP( ?x :p2 ?v2 ) ),
     *         BGP( ?y :p3 ?v2 ) )
     * coreference constraints against ?v2 should only be enforced for
     * tuples from the right side of this union.
     */
    val pathNo = sql.ProjectAttribute(sql.PrimaryExpressionTyped(rdb.RDB.Datatype.INTEGER,sql.Name("" + initState.joins.size)),
					sql.AttrAlias(sql.Name("_DISJOINT_")))
    val leftJoinVars = gp.findVars
    val attrlist:Set[sql.ProjectAttribute] = leftJoinVars.map(
      v =>
	sql.ProjectAttribute(varToAttribute(nestedState.varmap, sparql.VarAssignable(v)),
			   sql.AttrAlias(attrAliasNameFromVar(sparql.VarAssignable(v))))
    ) + pathNo // add join number to selections
    val subselect = sql.Select(
      false,
      sql.Projection(attrlist),
      sql.TableList(nestedState.joins),
      nestedState.exprs.size match {
      	case 0 => None
      	case 1 => Some(nestedState.exprs.toList(0))
      	case _ => Some(sql.ExprConjunction(nestedState.exprs))
      },
      List[sql.OrderElt](), None, None
    )

    /** Create a condition to test if this OPTIONAL was matched (called
     * _DISJOINT_ as OPTIONAL behaves pretty much like a disjunction).
     */
    val nestedCond = sql.RelationalExpressionNull(sql.PrimaryExpressionAttr(
      sql.RelVarAttr(leftJoinAlias, rdb.RDB.AttrName("_DISJOINT_"))))

    /** Bind variables to the attributes projected from the subselect;
     * handle corefs (equivalence with earlier bindings).
     */
    val outerState2 =
      nestedVars.foldLeft(
	R2RState(initState.joins,
		 initState.varmap,
		 Set[sql.Expression]()))((myState, v) => 
		   subselectVars(myState, sparql.VarAssignable(v), leftJoinAlias, nestedCond,
				 initState.varmap, nestedState.varmap, true))

    /** The final state includes the subselect as a join, the variables bound
     * to subselect projection, and no new expresssions. The expressions
     * derived from corefs are conditions for the LEFT OUTER JOIN.
     */
    val join = sql.LeftOuterJoin(sql.AliasedResource(sql.Subselect(subselect), leftJoinAlias), 
	 outerState2.exprs.size match {
	   case 0 =>
	     sql.RelationalExpressionEq(sql.PrimaryExpressionTyped(rdb.RDB.Datatype.INTEGER,sql.Name("1")),
					sql.PrimaryExpressionTyped(rdb.RDB.Datatype.INTEGER,sql.Name("1")))
	     /** Require corefs unless we have a leading OPTIONAL. */
	     // if (...)
	     // else
	     //   error ("Nested GP has no variables shared with its context; cowaredly refusing to join ON 1.")
	   case 1 => outerState2.exprs.toList(0)
	   case _ => sql.ExprConjunction(outerState2.exprs)
	 }
       )
    val exprs =
      if (negate) {
	initState.exprs + sql.RelationalExpressionNull(sql.PrimaryExpressionAttr(sql.RelVarAttr(leftJoinAlias, rdb.RDB.AttrName("_DISJOINT_"))))
      } else initState.exprs
    R2RState(initState.joins + join, outerState2.varmap, exprs)
  }

  /**
   * Recursively add the joins, variable mappings and constraints for an SQL query implementing a graph pattern.
   * @param db  database description.
   * @param state  initial set of joins, variable mappings and constraints.
   * @param gp  the SPARQL GraphPattern to represent as SQL.
   * @param enforceForeignKeys  if true, SPARQL triple patterns corresponding to foreign keys, e.g. <code>?who :hasParent ?parent</code>, generate a join on the referenced table.
   * @return  a new set of joins, variable mappings and constraints.
   * Per <a href="http://www.w3.org/TR/rdf-sparql-query/#sparqlQuery">definition of SPARQL Query</a>, SPARQL Graph Patterns are (Basic Graph Pattern, Join, LeftJoin, Filter, Union, Graph).
   * mapGraphPattern maps each of these to an SQL abstract query (which can then be serialized as SQL and executed).
   * 
   */
  def mapGraphPattern(db:rdb.RDB.Database, state:R2RState, gp:sparql.GraphPattern, enforceForeignKeys:Boolean):R2RState = {
    gp match {

      /** <a href="http://www.w3.org/TR/rdf-sparql-query/#defn_PatternInstanceMapping">Basic Graph Pattern Matching</a>()
       * @param triplepatterns  set of triple patterns. Premise: all triple patterns must resolve against the direct graph.
       * As { TP1, TP2 } == Join({ TP1 }, { TP2 }), we can view the bindOnPredicate function as partitioning triple patterns by the relvar they match.
       * This isn't observable in the SQL query as all the triple patterns in all the conjunctions contribute to the same query.
       */
      case sparql.TriplesBlock(triplepatterns) => {
	/** Examine each triple, updating the compilation state. */
	val state2 = triplepatterns.foldLeft(state)((incState,s) => bindOnPredicate(db, incState, s, enforceForeignKeys))

	/** NULLs in the database result in no triple in the Direct Graph.
	 * Enforce this by requiring that the SQL expression to which any SPARQL variable (Assignable) is bound is NOT NULL.
	 */
	val nullExprs = gp.findAssignables.foldLeft(Set[sql.Expression]())((s, vvar) => {
	  if (varToAttributeDisjoints(state2.varmap, vvar).size == 0)
	    /** Create a NOT NULL expression for each fully bound variable. */
	    s ++ Set(sql.RelationalExpressionNotNull(sql.PrimaryExpressionAttr(varToAttribute(state2.varmap, vvar))))
	  else
	    /** Variables in a partial binding can be NULL so the aren't added to the null expressions. */
	    s
	})
	R2RState(state2.joins, state2.varmap, state2.exprs ++ nullExprs)
      }

      /** <a href="http://www.w3.org/TR/rdf-sparql-query/#defn_evalJoin">Join</a>(P1, P2)
       * Since Join is association, we handle this as an n-ary join.
       * @param conjoints  list of graph patterns to join.
       */
      case sparql.TableConjunction(conjoints) => {
	conjoints.foldLeft(state)((incState,s) => mapGraphPattern(db, incState, s, enforceForeignKeys))
      }

      /** <a href="http://www.w3.org/TR/rdf-sparql-query/#defn_evalLeftJoin">LeftJoin</a>(P1, P2, F)
       * The parser providing the SPARQL abstract query turns LeftJoin(P1, P2, F) into Join(P1, Optional(P2)), or Join(P1, Optional(Filter(P2, F))) if there is a FILTER.
       * @param gp2  nested graph pattern (Ω in algebra)
       */
      case sparql.OptionalGraphPattern(gp2) => {
	/* state_postLeadingTable: create an initial table if the first conjoint is optional.
	 * e.g. ... FROM (SELECT 1 AS _EMPTY_) AS _EMPTY_ LEFT OUTER JOIN ...
	 */
	val state_postLeadingTable =
	  if (state.joins.size == 0)
	    /**
	     * Leading optionals (ASK WHERE { OPTIONAL { ... } ... }) in SPARQL don't have a counterpart in SQL.
	     * We emulate leading optionals with a leading SQL subselect which projects one solution with no selected attributes.
	     */
	    R2RState(state.joins + sql.InnerJoin(sql.AliasedResource(sql.Subselect(
	      sql.Select(
		false, 
		sql.Projection(Set(sql.ProjectAttribute(sql.PrimaryExpressionTyped(rdb.RDB.Datatype.INTEGER,sql.Name("1")),
							 sql.AttrAlias(sql.Name("_EMPTY_"))))),
		sql.TableList(util.AddOrderedSet()),
		None, List[sql.OrderElt](), None, None
	      )), sql.RelVar(sql.Name("_EMPTY_"))), None), state.varmap, state.exprs)
	  else
	    state
	/** Create an OUTER JOIN for the nested graph pattern. */
	synthesizeOuterJoin(state_postLeadingTable, gp2, false, db, enforceForeignKeys)
      }

      /** <a href="http://www.w3.org/TR/rdf-sparql-query/#defn_algFilter">Filter</a>(expr, Ω)
       * @param gp2  nested graph pattern (Ω in algebra)
       * @param expr  boolean SPARQL expression (expr in algebra)
       */
      case sparql.TableFilter(gp2:sparql.GraphPattern, expr:sparql.Expression) => {
	/** Calculate state for gp2. */
	val state2 = mapGraphPattern(db, state, gp2, enforceForeignKeys)

	/** Add constraints for all the FILTERS */
	val filterExprs:Set[sql.RelationalExpression] =
	  expr.conjuncts.toSet map ((x:sparql.PrimaryExpression) => filter2expr(state2.varmap, x))

	R2RState(state2.joins, state2.varmap, state2.exprs ++ filterExprs)
      }

      /** <a href="http://www.w3.org/TR/rdf-sparql-query/#defn_evalUnion">Union</a>(P1, P2)
       * Since Disjunction is associative, we handle this as an n-ary disjunction.
       * @param disjoinits  list of graph patterns to concatenate.
       */
      case sparql.TableDisjunction(disjoints) => {
	/**
	 * SPARQL UNIONs are realized as SQL subselects.
	 * Set up initial state for this subselect.
	 */
	val unionAlias = sql.RelVar(sql.Name("G_union" + state.joins.size)) // invent a unique name for this union.
	val emptyState = R2RState( // create an empty state.
	  util.AddOrderedSet[sql.Join](), 
	  Map[sparql.Assignable, SQL2RDFValueMapper](), 
	  Set[sql.Expression]()
	)
	val unionVars = disjoints.foldLeft(Set[sparql.Var]())((mySet,disjoint) =>
	  mySet ++ disjoint.findVars).toList // all variables nested in the disjoints.

	/**
	 * Map the list of disjoints to a list of nested R2RStates, nested variable lists, and unique SQL constants identifying that disjoint.
	 * Non-Functional var <code>number</code> is used for projecting unique
	 * constants to indicate which disjoint produced a tuple.
	 */
	var number = 0
	val nestedStates = disjoints.map(disjoint => {
	  val disjointState = mapGraphPattern(db, emptyState, disjoint, enforceForeignKeys)
	  val disjointVars = disjoint.findVars
	  val uniqueConst = sql.PrimaryExpressionTyped(rdb.RDB.Datatype.INTEGER,sql.Name("" + number))
	  number = number + 1 // non-functional, but clearer than wrapping as a parameter in a foldLeft
	  (disjointState, disjointVars, uniqueConst)
	})

	/**
	 * Map the list of nested R2RStates to a set of subselects.
	 * <code>uniqueConst</code> is used for projecting a value
	 * to indicate which disjoint produced a tuple.
	 */
	val subselects = nestedStates.foldLeft(Set[sql.Select]())((subselects, state) => {
	  val (disjointState, disjointVars, uniqueConst) = state
	  /**
	   * Select a constant as _DISJOINT_ so later constraints can be
	   * sensitive to whether a variable was bound.
	   * This matters for OPTIONALs and, here, assymetric UNIONs. Given:
	   *   Join( Union( BGP( ?x :p1 ?v1 ),
	   *                BGP( ?x :p2 ?v1 . ?x :p2 ?v2 ) ),
	   *         BGP( ?y :p3 ?v2 ) )
	   * coreference constraints against ?v2 should only be enforced for
	   * tuples from the right side of this union.
	   */
	  val pathNo = sql.ProjectAttribute(uniqueConst,
					  sql.AttrAlias(sql.Name("_DISJOINT_")))

	  val attrlist:Set[sql.ProjectAttribute] = unionVars.foldLeft(Set(pathNo))((attrs, v) => {
	    val attrOrNull = if (disjointState.varmap.contains(sparql.VarAssignable(v))) varToAttribute(disjointState.varmap, sparql.VarAssignable(v)) else sql.ConstNULL()
	    attrs ++ Set(sql.ProjectAttribute(attrOrNull, sql.AttrAlias(attrAliasNameFromVar(sparql.VarAssignable(v)))))
	  })

	  val subselect = sql.Select(
	    false,
	    sql.Projection(attrlist),
	    sql.TableList(disjointState.joins),
	    disjointState.exprs.size match {
	      case 0 => None
	      case 1 => Some(disjointState.exprs.toList(0))
	      case _ => Some(sql.ExprConjunction(disjointState.exprs))
	    }, List[sql.OrderElt](), None, None
	  )
	  subselects + subselect
	})

	/**
	 * Connect the variables projected from the nested selects into the outer variable bindings and constraints.
	 * <code>state2</code> will have no additional tables in the TableList.
	 * <code>uniqueConst</code> is used this time to constraint coreferences between the
	 * subselects and the outer context.
	 */
	val state2 = nestedStates.foldLeft(state)((outerState, state) => {
	  val (disjointState, disjointVars, uniqueConst) = state

	  /** Create a condition to test if this disjoint was matched. */
	  val disjointCond = sql.RelationalExpressionNe(sql.PrimaryExpressionAttr(sql.RelVarAttr(unionAlias, rdb.RDB.AttrName("_DISJOINT_"))),
							uniqueConst)
	  val outerState2 = disjointVars.foldLeft(outerState)((myState, v) =>
	      subselectVars(myState, sparql.VarAssignable(v), unionAlias, disjointCond, outerState.varmap, disjointState.varmap, false))
	  number = number + 1 // non-functional, but clearer than wrapping as a parameter in a foldLeft
	  outerState2
	})
	val subselect = sql.Subselect(sql.Union(subselects))
	R2RState(state.joins + sql.InnerJoin(sql.AliasedResource(subselect,unionAlias), None), state2.varmap, state2.exprs)
      }

      /** <a href="http://www.w3.org/TR/rdf-sparql-query/#defn_evalGraph">Graph</a>(IRI, P)
       * I don't know what the parser did with the IRI, but we don't know what to do with GRAPHs anyways. 
       * @param gp2  nested graph pattern (Ω in algebra)
       */
      case sparql.GraphGraphPattern(gp2) => error("no code to handle GraphGraphPatterns (" + gp2 + ")")

      /** Minus is from SPARQL 1.1 (in progress). This doesn't need documentation now.
       * @param gp2  the graph pattern to subtract.
       */
      case sparql.MinusGraphPattern(gp2) => {
	if (state.joins.size == 0) state
	else synthesizeOuterJoin(state, gp2, true, db, enforceForeignKeys)
      }
    }
  }

  /**
   * Default interface for SparqlToSql.
   * @param db  database description.
   * @param sparquery  SPARQL compile tree.
   * @param stem  stem URI for all generated RDF URIs.
   * @param enforceForeignKeys  if true, SPARQL triple patterns corresponding to foreign keys, e.g. ?who :hasParent ?parent , generate a join on the referenced table.
   * @param concat  if true, keys will produce SQL functions to generate a URI, e.g. SELECT CONCAT(stemURI, table, "/", pk, ".", R_who.pk) AS who
   * @return an SQL query corresponding to sparquery
   */
  def apply (db:rdb.RDB.Database, sparquery:sparql.Select, stem:StemURI, enforceForeignKeys:Boolean, concat:Boolean) : (sql.Select, Map[sparql.Assignable, SQL2RDFValueMapper]) = {

    /** Create an object to hold our compilation state. */
    val initState = R2RState(
      util.AddOrderedSet[sql.Join](), 
      Map[sparql.Assignable, SQL2RDFValueMapper](), 
      Set[sql.Expression]()
    )

    /**
     * Generate a new state with the joins, mappings to sql expressions, and
     * constraints implicit in the SPARQL WHERE pattern.
     */
    val r2rState = mapGraphPattern(db, initState, sparquery.gp, enforceForeignKeys)

    /**
     * Select the attributes corresponding to the variables
     * in the SPARQL SELECT.
     */
    val attrlist:Set[sql.ProjectAttribute] =
      // This foldLeft could be a map, if i could coerce to a set afterwards.
      sparquery.attrs.attributelist.foldLeft(Set[sql.ProjectAttribute]())((attrs, v) => {
	val exp =
	  if (concat)
	    // generate CONCAT expression for keys.
	    varToExpr(r2rState.varmap, sparql.VarAssignable(v), stem)
	  else
	    varToAttribute(r2rState.varmap, sparql.VarAssignable(v))
	/** Projection alias. */
	val as = sql.AttrAlias(attrAliasNameFromVar(sparql.VarAssignable(v)))
	attrs + sql.ProjectAttribute(exp , as)
      })

    /** Construct the generated query as an abstract syntax. */
    val select = sql.Select(
      sparquery.distinct,
      sql.Projection(attrlist),
      sql.TableList(r2rState.joins),
      r2rState.exprs.size match {
	case 0 => None
	case 1 => Some(r2rState.exprs.toList(0))
	case _ => Some(sql.ExprConjunction(r2rState.exprs))
      },
      sparquery.order.map((elt:sparql.OrderElt) => {
	sql.OrderElt(elt.desc, xxx(r2rState.varmap, elt.expr))
      }), sparquery.offset, sparquery.limit
    )
    // println("r2rState.varmap: " + r2rState.varmap)
    // println("select.expression: " + select.expression)
    (select.makePretty, r2rState.varmap)
  }

/*
 * vvv CLEAN THIS UP!!! vvv
 */

  def assignable2expr999(varmap:Map[sparql.Assignable, SQL2RDFValueMapper], rTerm:sparql.Term):sql.Expression = { // :sparql.Var
    val r:sql.PrimaryExpression = rTerm match {
      case sparql.TermUri(u) => error("not implemented: translating RDF URI to SQL: " + u) // :sparql.Uri
      case sparql.TermVar(v) => sql.PrimaryExpressionAttr(varToAttribute(varmap, sparql.VarAssignable(v)))
      case sparql.TermBNode(b) => sql.PrimaryExpressionAttr(varToAttribute(varmap, sparql.BNodeAssignable(b)))
      case sparql.TermLit(sparql.Literal(p:rdf.PlainLiteral)) => error("not implemented")
      case sparql.TermLit(sparql.Literal(rdf.TypedLiteral(lit,dt))) =>
	sql.PrimaryExpressionTyped({
	  dt match {
	    case rdf.IRI("http://www.w3.org/2001/XMLSchema#string") => rdb.RDB.Datatype.STRING
	    case rdf.IRI("http://www.w3.org/2001/XMLSchema#integer") => rdb.RDB.Datatype.INTEGER
	    case rdf.IRI("http://www.w3.org/2001/XMLSchema#date") => rdb.RDB.Datatype.DATE
	    case _ => error("unable to translate to RDF literal SQL: \"" + lit + "\"^^<" + dt + ">")
	  }
	}, lit)
    }
    r
  }

  def xxx(varmap:Map[sparql.Assignable, SQL2RDFValueMapper], from:sparql.Expression) : sql.Expression = {
    val l = from.conjuncts.map((conj) => {
      conj match {
	case sparql.SparqlTermExpression(sparql.TermVar(v:sparql.Var)) =>
	  assignable2expr999(varmap, sparql.TermVar(v))
	case sparql.SparqlTermExpression(sparql.TermBNode(b:sparql.BNode)) =>
	  assignable2expr999(varmap, sparql.TermBNode(b))
	case sparql.SparqlTermExpression(sparql.TermLit(l)) =>
	  assignable2expr999(varmap, sparql.TermLit(l))
	case sparql.SparqlTermExpression(sparql.TermUri(u:sparql.Uri)) =>
	  assignable2expr999(varmap, sparql.TermUri(u))
	case e:sparql.PrimaryExpression => yyy(varmap, e)
      }})
    if (l.size == 1)
      l(0)
    else
      sql.ExprConjunction(l.toSet)
  }

  def yyy(varmap:Map[sparql.Assignable, SQL2RDFValueMapper], from:sparql.PrimaryExpression) : sql.Expression = {
    from match {
      case sparql.SparqlTermExpression(term) => assignable2expr999(varmap, term)
      case sparql.PrimaryExpressionEq(l, r) => sql.RelationalExpressionEq(yyy(varmap, l), yyy(varmap, r))
      case sparql.PrimaryExpressionGt(l, r) => sql.RelationalExpressionEq(yyy(varmap, l), yyy(varmap, r))
      case sparql.PrimaryExpressionLt(l, r) => sql.RelationalExpressionEq(yyy(varmap, l), yyy(varmap, r))
    }
  }

  /**
   * junk that should be elsewhere
   */

  implicit def relname2relresource (rn:rdb.RDB.RelName) : sql.RelationResource = sql.RelationResource(rn)

}

/**
 * Support functions to inject SparqlToSql results into an XML Results Set.
 * @example:
 * <pre>
 * val xmlres:String = 
 *   head(List[String]("emp", "name")) +
 *   startresult +
 *     binding("emp", "253", rdfmap, stem) +
 *     binding("name", "Bob", rdfmap, stem) +
 *   endresult +
 *   startresult +
 *     binding("emp", "258", rdfmap, stem) +
 *     // employee 258 has no name attribute so omit this binding
 *   endresult +
 *   foot
 * </pre>
 *
 * @see {@link org.w3.sparql2sql.SparqlToSql}
 * @see {@link http://www.w3.org/TR/2008/REC-rdf-sparql-XMLres-20080115/ XML Results Format}
 */
object SqlToXMLRes {

  /**
   * Create a SPARQL Results format header and begin the body (results).
   * @param vars list of variable names to insert into the header
   */
  def head (vars:List[String]) : String = {
    "<?xml version=\"1.0\"?>\n<sparql xmlns=\"http://www.w3.org/2005/sparql-results#\">\n  <head>\n" +
    vars.map(varname => "    <variable name=\"" + varname + "\"/>\n").mkString + 
    "  </head>\n\n  <results>\n"
  }

  /**
   * Open a result element
   */
  def startresult () : String = {
    "    <result> \n"
  }

  /**
   * Create a binding value appropriate for <code>name</code>'s datatype.
   * @param name name of bound variable.
   * @param value lexical value of bound variable, may need normalization from e.g. SQL.
   * @param varmap mapping of sparql variables to datatypes, emitted by SparqlToSql._2
   * @param stem  stem URI for all generated RDF URIs.
   */
  def binding (name:String, value:String, varmap:Map[sparql.Assignable, SparqlToSql.SQL2RDFValueMapper], stem:StemURI) : String = {
    def getattr (b:SparqlToSql.FullOrPartialBinding) : rdb.RDB.AttrName = {
      b match {
	case SparqlToSql.FullBinding(sql.RelVarAttr(_, attr)) =>
	  attr
	case SparqlToSql.PartialBinding(binders) => {
	  val SparqlToSql.BindingConstraint(expr, sql.RelVarAttr(rv, attr)) = binders.toList(0)
	  attr
	}
      }
    }
    val t:String = varmap(sparql.VarAssignable(sparql.Var(name))) match {
      case SparqlToSql.RDFNoder(rel, b)  => "<uri>" + stem.s + rel + "/" + getattr(b) + "." + value + "#record</uri>"
      case SparqlToSql.RDFBNoder(rel, b) => "<bnode>bnode_" + rel + "/" + getattr(b) + "." + value + "</bnode>"
      case SparqlToSql.DateMapper(_)     => "<literal datatype=\"http://www.w3.org/2001/XMLSchema#date\">" + value + "</literal>"
      case SparqlToSql.IntMapper(_)      => "<literal datatype=\"http://www.w3.org/2001/XMLSchema#integer\">" + value + "</literal>"
      case SparqlToSql.StringMapper(_)   => "<literal datatype=\"http://www.w3.org/2001/XMLSchema#string\">" + value + "</literal>"
    }
    "      <binding name=\"" + name + "\">\n	" + t + "\n      </binding>\n"
  }

  /**
   * Close a result element.
   */
  def endresult () : String = {
    "    </result>\n"
  }

  /**
   * End SPARQL Results document.
   */
  def foot () : String = {
    "  </results>\n</sparql>\n"
  }

}

