package org.w3.util.fig

import io.{Source, Codec}
import java.io.File
import net.liftweb.json._
import net.liftweb.json.JsonAST._

// adapted from http://github.com/codahale/fig

object JsonBasedConfiguration {
  /**
   * returns the content of a file, if accessible from the classpath
   */
  def fromFileInClasspath(filename:String):String = {
    val inputStream = this.getClass.getResourceAsStream("/"+filename)
    Source.fromInputStream(inputStream)(Codec.UTF8).getLines().mkString("\n")
  }

  /**
   * maps the value to an Option. null or exception give None
   */
  def option[T](value : => T) : Option[T] = {
    try {
      Option(value)
    } catch {
      case _ => None
    }
  }

}

/**
 * An exception class thrown when there is a configuration error.
 */
class ConfigurationException(message: String) extends Exception(message)

/**
 * A JSON-based configuration file. Line comments (i.e., //) are allowed.
 *
 * val config = new Configuration("config.json")
 * config("rabbitmq.queue.name").as[String]
 *
 * @author coda
 */
class JsonBasedConfiguration(filename: String) {
  case class Value(path: String, value: JsonAST.JValue) {
    /**
     * Returns the value as an instance of type A.
     */
    def as[A](implicit mf: Manifest[A]) = value.extract[A](DefaultFormats, mf)

    /**
     * Returns the value as an instance of type Option[A]. If the value exists,
     * Some(v: A) is returned; otherwise, None.
     */
    def asOption[A](implicit mf: Manifest[A]) = value.extractOpt[A](DefaultFormats, mf)

    /**
     * Returns the value as an instance of type A, or if the value does not
     * exist, the result of the provided function.
     */
    def or[A](default: => A)(implicit mf: Manifest[A]) = asOption[A](mf).getOrElse(default)

    /**
     * Returns the value as an instance of type A, or if it cannot be converted,
     * throws a ConfigurationException with an information error message.
     */
    def asRequired[A](implicit mf: Manifest[A]) = asOption[A] match {
      case Some(v) => v
      case None => throw new ConfigurationException(
        "%s property %s not found".format(mf.erasure.getSimpleName, path)
      )
    }

    /**
     * Returns the value as a instance of List[A], or if the value is not a JSON
     * array, an empty list.
     */
    def asList[A](implicit mf: Manifest[A]) = value match {
      case JField(_, JArray(list)) => list.map { _.extract[A](DefaultFormats, mf) }
      case other => List()
    }

    /**
     * Returns the value as an instance of Map[String, A], or if the value is
     * not a simple JSON object, an empty map.
     */
    def asMap[A](implicit mf: Manifest[A]) = value match {
      case JField(_, o: JObject) =>
        o.obj.map { f => f.name -> f.value.extract[A](DefaultFormats, mf) }.toMap
      case other => Map()
    }
  }

  private val json = {
    val content = JsonBasedConfiguration.option(Source.fromFile(new File(filename)).mkString) getOrElse JsonBasedConfiguration.fromFileInClasspath(filename)
    JsonParser.parse(content.replaceAll("""(^//.*|[\s]+//.*)""", ""))
  }

  /**
   * returns all the available keys from the configuration file
   */
  val availableKeys:List[String] = {
    def getKeys(json:JsonAST.JValue):List[String] = {
      json match {
        case JsonAST.JField(name, value) if value.children.size == 0 =>
          List(name)
        case JsonAST.JField(name, value) =>
          value.children flatMap { getKeys(_) } map { name + "." + _ }
        case _ => List()
      }
    }
    json.children flatMap { getKeys(_) }
  }
  
  /**
   * Given a dot-notation JSON path (e.g., "parent.child.fieldname"), returns
   * a Value which can be converted into a specific type or Option thereof.
   */
  def apply(path: String) = Value(path, path.split('.').foldLeft(json) { _ \ _ })
}
