package org.w3.sparql2sqlendpoint

import javax.servlet.http.{HttpServlet, HttpServletRequest, HttpServletResponse} 
import scala.xml.XML

import java.sql.{DriverManager, Connection, Statement, ResultSet, ResultSetMetaData}

import java.sql.{Types => T}

import java.net.URI
import org.w3.rdb.RDB.{RelName,AttrName}
import org.w3.sql.SqlParser
import org.w3.sparql
import org.w3.sparql.Sparql
import org.w3.sql
import org.w3.sparql2sql.{SparqlToSql,StemURI,SqlToXMLRes}

object Control {

  private def using[Closeable <: {def close(): Unit}, B](closeable: Closeable)(getB: Closeable => B): B =
    try {
      getB(closeable)
    } finally {
      closeable.close()
    }

  private def bmap[T](test: => Boolean)(block: => T): List[T] = {
    val ret = new scala.collection.mutable.ListBuffer[T]
    while(test) ret += block
    ret.toList
  }

  /** Executes the SQL and processes the result set using the specified function. */
  private def query[B](connection: Connection, sql: String)(process: ResultSet => B): B =
    try {
      using (connection) { connection =>
        using (connection.createStatement) { statement =>
          using (statement.executeQuery(sql)) { results =>
            process(results)
          }
        }
      }
    } catch {
      case e => throw new Exception("Malformed SQL query: \n" + sql, e)
    }

  /** Executes the SQL and uses the process function to convert each row into a T. */
  def queryEach[T](connection: Connection, sql: String)(process: ResultSet => T): List[T] =
    query(connection, sql) { results =>
      bmap(results.next) {
        process(results)
      }
    }

  def process(connection: Connection, sql: String,
              head: List[String] => String,
              startres: String,
          binding: (String, String /*, Map[sparql.Assignable, SparqlToSql.SQL2RDFValueMapper], StemURI*/) => String,
          endres: String, foot: String): String = {

    query(connection, sql) { results =>

      val metadata = results.getMetaData

      val vars:List[String] =
        (1 to metadata.getColumnCount) map { column => metadata.getColumnLabel(column) } toList

      def processCell(rs:ResultSet, column:Int):Option[String] =
	try {
          val name:String = metadata.getColumnLabel(column)
          val value:String =
	    metadata.getColumnType(column) match {
              case T.VARCHAR => rs.getString(column)
              case T.INTEGER => rs.getInt(column).toString
              case T.DATE    => rs.getDate(column).toString
              case _         => throw new java.lang.Exception("you have to map this type!!!")
            }
          Some(binding(name, value))
	} catch {
	  case _ => None
	}

      def processRow(rs:ResultSet):String =
        (1 to metadata.getColumnCount) flatMap { column => processCell(rs, column) } mkString ""

      val rows:List[String] = bmap(results.next) {
        processRow(results)
      }

      val body = rows map { startres + _ + endres } mkString ""

      head(vars) + body + foot

    }

  }

}

import org.w3.util.fig._

trait Configuration {
  def getConfigurationValue(key:String):String
  def availableKeys:List[String]
}

trait DefaultJsonConfiguration extends Configuration {

  val config = new JsonBasedConfiguration("default-config.json")

  def getConfigurationValue(key:String) = config(key).as[String]

  val availableKeys = config.availableKeys

  Class.forName(getConfigurationValue("db.driver")).newInstance

  def getConnection():Connection =
    DriverManager.getConnection(getConfigurationValue("db.url"),
                                getConfigurationValue("db.user"),
                                getConfigurationValue("db.password"))
  
  val defaultStemURI:String = getConfigurationValue("ui.stemuri")

  val defaultSparqlQuery:String = getConfigurationValue("ui.sparql-query")

  val db:org.w3.rdb.RDB.Database = {
    val DDLParser = new org.w3.sql.SqlParser()
    val ddl = getConfigurationValue("ddl")
    DDLParser.parseAll(DDLParser.ddl, ddl).get
  }

}

class SPARQL2SQLEndpoint extends HttpServlet with DefaultJsonConfiguration {

  val encoding = "utf-8"

  override def doGet(request:HttpServletRequest, response:HttpServletResponse) {

    request.getParameter("query") match {
      case null | "" => processIndex(request, response)
      case query     => processSparql(request, response, query)
    }

  }

  def processSparql(request:HttpServletRequest, response:HttpServletResponse, query:String) {

    val stemURI = new StemURI(Some(request.getParameter("stemuri")) getOrElse defaultStemURI)

    val sparqlParser = Sparql()

    val sparqlSelect = sparqlParser.parseAll(sparqlParser.select, query).get

    val (generated, rdfmap) = SparqlToSql(db, sparqlSelect, stemURI, true, false)

    val connection:Connection = getConnection()
    // if (! connection.isClosed) println("Successfully connected")

    val res = Control.process(connection=connection,
			      sql=generated.toString,
			      head=SqlToXMLRes.head(_),
			      startres=SqlToXMLRes.startresult,
			      binding=SqlToXMLRes.binding(_, _, rdfmap, stemURI),
			      endres=SqlToXMLRes.endresult,
			      foot=SqlToXMLRes.foot)

    response.setContentType("text/plain; charset='" + encoding + "'")

    response.getWriter.write(res)

  }

  def processIndex(request:HttpServletRequest, response:HttpServletResponse) {

    val title = "SPARQL endpoint for the Direct Mapping"

    val index =
      <html xmlns="http://www.w3.org/1999/xhtml">
        <head><title>{ title }</title></head>
        <body>
          <h1>Query the <a href="http://www.w3.org/2001/sw/rdb2rdf/directGraph/">Direct Graph</a> using SPARQL!</h1>
          <form action="sparql">
            <p>
              StemURI: <input cols="80" name="stemuri" id="stemuri" value={ defaultStemURI } /><br />
              <textarea rows="10" cols="80" name="query" id="query">{ defaultSparqlQuery }</textarea>
              <input type="submit" />
            </p>
          </form>
          <hr />
          <address>
            <a href="http://www.w3.org/People/Eric/">Eric Prud'hommeaux</a>, <a href="http://www.w3.org/People/Bertails/">Alexandre Bertails</a>, Nov 2010
          </address>
        </body>
      </html>

    response.setContentType("application/xml; charset='" + encoding + "'")

    XML.write(response.getWriter, index, encoding, false, null) 

  }

}

object SPARQL2SQLEndpoint {

  import org.eclipse.jetty.server.Server
  import org.eclipse.jetty.servlet.{ServletContextHandler, ServletHolder}

  /**
   * returns the value associated to a key
   * example: javar -jar mybundle.jar -db.user root
   */
  def getElementAfter(args: List[String], key:String):Option[String] = {
    args match {
      case head :: tail if head == key => tail.headOption
      case head :: tail => getElementAfter(tail, key)
      case Nil => None
    }
  }

  def main(args: Array[String]) {

    def createStatements(connection: => Connection) = {
      import Control._

      val tables = queryEach(connection, "SHOW TABLES;") { row => row.getNString(1) }

      val create = tables flatMap { tableName =>
        val showCreateTable = "SHOW CREATE TABLE " + tableName + ";"
        queryEach(connection, showCreateTable) { row => row.getString(2) }
      }

      /* remove the stuff the SQL parser does not understand yet */
      val createStatements = create map { _.replaceAll("\\).*?$", ")")
                                           .replaceAll(" KEY", "")
                                           .replaceAll("`", "")
                                           .replaceAll(" DEFAULT .*?,", ",")
                                           .replaceAll("`", "")
                                           .replaceAll(" NOT NULL", "") } mkString "\n"
      createStatements
    }

    val endpoint = new SPARQL2SQLEndpoint() {
      override def getConfigurationValue(key:String) = {
        val value = getElementAfter(args.toList, "-"+key)
        if (key == "ddl" && value == Some("auto")) {
          val statements = createStatements(getConnection)
          println("Found schema:\n" + statements)
          statements
        } else {
          value getOrElse super.getConfigurationValue(key)
        }
      }
    }

    val message = endpoint.availableKeys map { key => "* " + key } mkString(start="You can override the following parameters by adding -<parameter> <value> on the command line:\n",
                                                                            sep="\n",
                                                                            end="")
    println(message)

    /* see http://wiki.eclipse.org/Jetty/Tutorial/Embedding_Jetty */
    val server:Server = new Server(8080)
    val context:ServletContextHandler = new ServletContextHandler(ServletContextHandler.SESSIONS)
    context.setContextPath("/")
    server.setHandler(context)
    
    context.addServlet(new ServletHolder(endpoint),"/*");
    server.start
    server.join

  }

}
