package org.w3.sql

import org.w3.util._
import org.w3.rdb.RDB

import scala.util.parsing.combinator._

object SQLParsers extends RegexParsers {

  val int = """[0-9]+""".r
  val float = """([0-9]+)?\.([0-9]+)?""".r
  val dquote = "\"([^\"\\\\\n\r]|\\\\[tbnrf\\\"'])*\"".r
  val squote = "\'([^\'\\\\\n\r]|\\\\[tbnrf\\\''])*\'".r
}

import SQLParsers._

import scala.util.parsing.combinator._
import java.net.URI

sealed abstract class RelationORSubselect
case class Subselect(sel:SelectORUnion) extends RelationORSubselect {
  override def toString = "(\n       " + sel.toString.replace("\n", "\n       ") + "\n                  )"
}
sealed abstract class SelectORUnion
case class Select(distinct:Boolean, projection:Projection, tablelist:TableList, expression:Option[Expression], order:List[OrderElt], offset:Option[Int], limit:Option[Int]) extends SelectORUnion {
  override def toString = 
    "SELECT "+
    { if (distinct) "DISTINCT " else "" }+
    projection+"\n"+
    tablelist+
    { if (expression.isDefined) {" WHERE " + expression.get + " "} else "" }+
    { if (order.size > 0) {order.map(o => o.toString).mkString("ORDER BY ", " ", " ") } else "" }+
    { if (offset.isDefined) {"OFFSET " + offset + " "} else "" }+
    { if (limit.isDefined) {"LIMIT " + limit + " "} else "" }
}
case class RelationResource(rn:RDB.RelName) extends RelationORSubselect {
  override def toString = rn.toString
}
object foo { // doesn't work
  implicit def relname2relationref (rn:RDB.RelName) = RelationResource(rn)
}
case class Union(disjoints:Set[Select]) extends SelectORUnion {
  override def toString = "  " + (disjoints.toList.map(s => s.toString.replace("\n", "\n  ")).mkString("\nUNION\n  "))
}
case class Projection(attributes:Set[ProjectAttribute]) {
  // foo, bar
  override def toString = attributes.toList.sortWith((l, r) => l.attralias.toString < r.attralias.toString).mkString(", ")
}
case class ProjectAttribute(value:RelVarAttrORExpression, attralias:AttrAlias) {
  override def toString = value + " AS " + attralias
}
//case class RelAttribute(relation:Relation, attribute:RDB.AttrName) c.f. ForeignKey999
sealed abstract class RelVarAttrORExpression
case class RelVarAttr(relvar:RelVar, attribute:RDB.AttrName) extends RelVarAttrORExpression {
  override def toString = relvar + "." + attribute
}

case class AttrAlias(n:Name) {
  override def toString = n.s /* "'" + n.s + "'" */
}
case class RelVar(n:Name) {
  override def toString = n.s /* "'" + n.s + "'" */
}
case class TableList(joins:AddOrderedSet[Join]) {
  override def toString =
    joins map { entry =>
      if (entry == joins.list(0)) entry
      else entry.joinText + " " + entry
    } mkString ("  FROM ", "\n       ", "")
}

sealed abstract class Join(res:AliasedResource) {
  override def toString = res.toString()
  def joinText():String
}
case class InnerJoin(res:AliasedResource, optOn:Option[Expression]) extends Join(res) {
  override def joinText () = "INNER JOIN";
}
case class LeftOuterJoin(res:AliasedResource, on:Expression) extends Join(res) {
  override def joinText () = "LEFT OUTER JOIN";
}

case class AliasedResource(rel:RelationORSubselect, as:RelVar) {
  override def toString = rel + " AS " + as
}
sealed abstract class Expression extends RelVarAttrORExpression
case class ExprConjunction(exprs:Set[Expression]) extends Expression {
  override def toString = "(" + (exprs.toList.sortWith((l, r) => l.toString < r.toString).mkString (")\n   AND (")) + ")"
}
case class ExprDisjunction(exprs:Set[Expression]) extends Expression {
  override def toString = "(" + (exprs mkString (") OR (")) + ")"
}
sealed abstract class RelationalExpression extends Expression
case class RelationalExpressionEq(l:Expression, r:Expression) extends RelationalExpression {
  override def toString = l + "=" + r
  /* safer operator== , but doesn't quite work yet. */
  // override def hashCode = 41 * l.hashCode + r.hashCode
  // override def equals(other: Any) = other match {
  //   case that: RelationalExpressionEq =>
  //     (that canEqual this) &&
  //   ( ( (this.l == that.l) && (this.r == that.r) ||
  //      (this.l == that.r) && (this.r == that.l) ) )
  //   case _ =>
  //     false
  // }
  // override def canEqual(other: Any) =
  //   other.isInstanceOf[RelationalExpressionEq]

  override def equals(that:Any) =
    that match {
      case RelationalExpressionEq(l1, r1) => (l == l1 && r == r1) || (r == l1 && l == r1)
      case _ => false
    }
  override def hashCode =
    if (r.hashCode < l.hashCode)
      r.hashCode + l.hashCode
    else
      l.hashCode + r.hashCode
}
case class RelationalExpressionNe(l:Expression, r:Expression) extends RelationalExpression {
  override def toString = l + "!=" + r
}
case class RelationalExpressionLt(l:Expression, r:Expression) extends RelationalExpression {
  override def toString = l + "<" + r
}
case class RelationalExpressionGt(l:Expression, r:Expression) extends RelationalExpression {
  override def toString = l + ">" + r
}
case class RelationalExpressionNull(l:Expression) extends RelationalExpression { // Expression?
  override def toString = l + " IS NULL"
}
case class RelationalExpressionNotNull(l:Expression) extends RelationalExpression { // Expression?
  override def toString = l + " IS NOT NULL"
}
sealed abstract class PrimaryExpression extends Expression
case class PrimaryExpressionAttr(fqattribute:RelVarAttr) extends PrimaryExpression {
  override def toString = "" + fqattribute
}
case class PrimaryExpressionTyped(datatype:RDB.Datatype, i:Name) extends PrimaryExpression {
  override def toString = /* "'" + i.s + "'" */ /* + datatype */
    datatype match {
      case RDB.Datatype("Int") => i.s
      case _ => "\"" + i.s + "\""
    }
}
case class OrderElt(desc:Boolean, expr:Expression) {
  override def toString = { if (desc) "DESC" else "ASC" } + "(" + expr.toString + ")"
}
case class ConstNULL() extends PrimaryExpression {
  override def toString = "NULL"
}
case class Concat(args:List[Expression]) extends PrimaryExpression {
  override def toString = args.mkString("CONCAT(", ", ", ")")
}
case class IfElse(cond:Expression, pass:Expression, fail:Expression) extends PrimaryExpression {
  override def toString = "CONCAT(" + cond + ", " + pass +  ", " + fail + ")"
}

case class Name(s:String)

object Name {
  implicit def fromStringToName(s:String):Name = Name(s)
}

sealed abstract class FieldDescOrKeyDeclaration
case class FieldDesc(attr:RDB.AttrName, datatype:RDB.Datatype, pkness:Boolean) extends FieldDescOrKeyDeclaration
sealed abstract class KeyDeclaration extends FieldDescOrKeyDeclaration
case class PrimaryKeyDeclaration(key:RDB.CandidateKey) extends KeyDeclaration
case class CandidateKeyDeclaration(key:RDB.CandidateKey) extends KeyDeclaration
case class ForeignKeyDeclaration(fk:List[RDB.AttrName], rel:RDB.RelName, pk:RDB.CandidateKey) extends KeyDeclaration
case class View(rel:RDB.RelName, defn:SelectORUnion) { // sibling of RDB.Relation
  override def toString = "CREATE VIEW " + rel + " AS\n" + defn
}

case class SqlParser() extends JavaTokenParsers {

  def toDB(s:String):RDB.Database = parseAll(ddl, s).get

  def toDB(file:java.io.File):RDB.Database = {
    val s = scala.io.Source.fromFile(file).getLines.reduceLeft(_+_)
    parseAll(ddl, s).get
  }

  def createview:Parser[View] = // @@@ could stick under ddl
    "(?i)CREATE".r ~ "(?i)VIEW".r ~ relation ~ "(?i)AS".r ~ selectORunion ^^
  { case _~_~relation~_~defn => View(relation, defn) }

  sealed abstract class DataDefinition
  case class Create(rn:RDB.RelName, relation:RDB.Relation) extends DataDefinition
  case class Insert(rn:RDB.RelName, tuple:RDB.Tuple) extends DataDefinition

  def ddl:Parser[RDB.Database] =
    rep1sep(createorinsert, ";") ~ opt(";") ^^
  {
    case l~x => RDB.Database(l.foldLeft(Map[RDB.RelName, RDB.Relation]())((m, p) => {
      p match {
	case Create(rn:RDB.RelName, relation:RDB.Relation) => m + (rn -> relation)
	case Insert(rn:RDB.RelName, tuple:RDB.Tuple) => m + (rn -> (m(rn) + tuple)) // add the tuple
      }
    }))
  }

  def createorinsert:Parser[DataDefinition] = (
      createtable ^^ { case c => c }
    | inserttable ^^ { case c => c }
  )

  def createtable:Parser[Create] =
    "(?i)CREATE".r ~ "(?i)TABLE".r ~ relation ~ "(" ~ rep1sep(fielddescorkeydef, ",") ~ ")" ^^
  {
    case _~_~relation~"("~reldesc~")" => {
      val pk0:Option[RDB.CandidateKey] = None
      val attrs0 = Map[RDB.AttrName, RDB.Datatype]()
      val candidates0 = List[RDB.CandidateKey]()
      val fks0 = Set[RDB.ForeignKey]()
      /* <pk>: (most recently parsed) PRIMARY KEY
       * <attrs>: map of attribute to type (e.g. INTEGER)
       * <fks>: map holding FOREIGN KEY relation REFERENCES attr
       */
      val (pk, attrs, candidates, fks) =
	reldesc.foldLeft((pk0, attrs0, candidates0, fks0)) { case ((pkopt, attrs, candidates, fks), rd) => {
	  rd match {
	    case FieldDesc(attr, value, pkness) => {
	      val (pkNew, candNew) =
		if (pkness) (Some(RDB.CandidateKey(List(attr))), candidates ++ List(RDB.CandidateKey(attr.n)))
		else (pkopt, candidates)
	      (pkNew, attrs + (attr -> value), candNew, fks)
	    }
	    case PrimaryKeyDeclaration(key) =>
	      // @@ why doesn't [[ candidates + RDB.CandidateKey(attr.n) ]] work?
	      (Some(key), attrs, candidates :+ key, fks)
	    case CandidateKeyDeclaration(key) =>
	      // @@ why doesn't [[ candidates + RDB.CandidateKey(attr.n) ]] work?
	      (pkopt, attrs, candidates :+ key, fks)
	    case ForeignKeyDeclaration(fk, rel, pk) =>
	      (pkopt, attrs, candidates, fks + RDB.ForeignKey(fk, RDB.Target(rel, pk)))
	  }
	} }
      val rd = RDB.Relation(relation, RDB.Header(attrs), List(), candidates, pk, RDB.ForeignKeys(fks))
      Create(relation, rd)
    }
  }

  def inserttable:Parser[Insert] =
    "(?i)INSERT".r ~ "(?i)INTO".r ~ relation ~ "(" ~ rep1sep(attribute, ",") ~ ")" ~ "(?i)VALUES".r ~ "(" ~ rep1sep(value, ",") ~ ")" ^^
  {
    case _~_~relation~"("~relnames~")"~_~"("~values~")" => {
      if (relnames.size != values.size)
	error("different numbers of elements in names " + relnames + " vs. values " + values)
      Map.empty ++ (relnames zip values)
      Insert(relation, RDB.Tuple(Map.empty ++ (relnames zip values)))
    }
  }

  def fielddescorkeydef:Parser[FieldDescOrKeyDeclaration] = (
      attribute ~ typpe ~ opt("(?i)PRIMARY".r ~ "(?i)KEY".r) ^^
      { case attribute~typpe~pkness => FieldDesc(attribute, typpe, pkness.isDefined) }
    | "(?i)PRIMARY".r ~ "(?i)KEY".r ~ "(" ~ rep1sep(attribute, ",") ~ ")" ^^
      { case _~_~"("~attributes~")" => PrimaryKeyDeclaration(RDB.CandidateKey(attributes)) }
    | "(?i)UNIQUE".r ~ "(" ~ rep1sep(attribute, ",") ~ ")" ^^
      { case _~"("~attributes~")" => CandidateKeyDeclaration(RDB.CandidateKey(attributes)) }
    | "(?i)FOREIGN".r ~ "(?i)KEY".r ~ "(" ~ rep1sep(attribute, ",") ~ ")" ~ "(?i)REFERENCES".r ~ relation ~ "(" ~ rep1sep(attribute, ",") ~ ")" ^^
      { case _~_~"("~fk~")"~_~relation~"("~pk~")" => ForeignKeyDeclaration(fk, relation, RDB.CandidateKey(pk)) }  )

  def typpe:Parser[RDB.Datatype] = (
      "(?i)INTEGER".r ~ opt(size)^^ { case _ => RDB.Datatype.INTEGER }
    | "(?i)INT".r ~ opt(size) ^^ { case _ => RDB.Datatype.INTEGER } // alias for INTEGER
    | "(?i)DOUBLE".r ^^ { case _ => RDB.Datatype.DOUBLE }
    | "(?i)FLOAT".r ^^ { case _ => RDB.Datatype.FLOAT }
    | "(?i)STRING".r ^^ { case _ => RDB.Datatype.STRING }
    | "(?i)CHAR".r ~ opt(size) ^^ { case _ => RDB.Datatype.STRING }
    | "(?i)VARCHAR".r ~ opt(size) ^^ { case _ => RDB.Datatype.STRING }
    | "(?i)DATETIME".r ^^ { case _ => RDB.Datatype.DATETIME }
    | "(?i)DATE".r ^^ { case _ => RDB.Datatype.DATE }
  )

  def size:Parser[Int] =
    "(" ~ int ~ ")" ^^ { case l ~ s ~ r => s.toInt }

  def selectORunion:Parser[SelectORUnion] =
    rep1sep(select, "(?i)UNION".r) ^^ { l => if (l.size == 1) l(0) else Union(l.toSet) }

  def select:Parser[Select] =
    "(?i)SELECT".r ~ opt("(?i)DISTINCT".r) ~ projection ~ opt(tablelist) ~ opt(where) ~ opt(order) ~ opt(offset) ~ opt(limit) ^^
    {
      case _ ~ distinct ~ attributes ~ tablesANDons ~ whereexpr ~ order ~ offset ~ limit => {
	val (tables, onExpressions) =
	  tablesANDons.getOrElse((TableList(AddOrderedSet[Join]()), Set[Expression]()))
	val t:Set[Expression] = onExpressions
	val onConjoints:Set[Expression] =
	  onExpressions.foldLeft(Set[Expression]())((s, ent) =>
	  s ++ {ent match {
	    case ExprConjunction(l) => l
	    case _ => Set(ent)
	  }})
	val conjoints = whereexpr match {
	  case Some(ExprConjunction(l)) => onConjoints ++ l
	  case Some(x) => onConjoints + x
	  case _ => onConjoints
	}
	val expr:Option[Expression] = conjoints.size match {
	  case 0 => None
	  case 1 => Some(conjoints.toList(0))
	  case _ => Some(ExprConjunction(conjoints))
	}
	Select(distinct.isDefined, attributes, tables, expr, order.getOrElse(List[OrderElt]()), offset, limit)
      }
    }

  def order:Parser[List[OrderElt]] = 
    "(?i)ORDER".r ~ "(?i)BY".r ~ rep(orderelt) ^^ { case o~b~elts => elts }

  def orderelt:Parser[OrderElt] = (
      "(?i)ASC".r ~ "(" ~ expression ~ ")" ^^ { case a~o~expr~c => OrderElt(false, expr) }
    | "(?i)DESC".r ~ "(" ~ expression ~ ")" ^^ { case a~o~expr~c => OrderElt(true, expr) }
    | fqattribute ^^ { case v => OrderElt(false, PrimaryExpressionAttr(v)) }
  )
    
  def offset:Parser[Int] =
    "(?i)OFFSET".r ~ int ^^ { case o~i => i.toInt }

  def limit:Parser[Int] =
    "(?i)LIMIT".r ~ int ^^ { case o~i => i.toInt }

  def where:Parser[Expression] =
    "(?i)WHERE".r ~ expression ^^ { case _ ~ expression => expression }

  def projection:Parser[Projection] =
    repsep(namedattribute, ",") ^^ { l => Projection(l.toSet) }

  def namedattribute:Parser[ProjectAttribute] =
    fqattributeORprimaryexpression ~ "(?i)AS".r ~ attralias ^^
    { case fqattributeORprimaryexpression ~ _ ~ attralias =>
      ProjectAttribute(fqattributeORprimaryexpression, attralias) }

  def fqattributeORprimaryexpression:Parser[RelVarAttrORExpression] = (
      fqattribute ^^ { case fqattribute => fqattribute }
    | primaryexpression ^^ { case const => const }
  )

  def fqattribute:Parser[RelVarAttr] =
    relvar ~ "." ~ attribute ^^
    { case relvar ~ "." ~ attribute => RelVarAttr(relvar, attribute) }

  def attribute:Parser[RDB.AttrName] = (
      """[a-zA-Z_]\w*""".r ^^ { x => RDB.AttrName(x) }
    | """`[^`]+`""".r ^^ { x => RDB.AttrName(x.substring(1, x.size - 1)) }
  )

  def attralias:Parser[AttrAlias] = (
      """[a-zA-Z_]\w*""".r ^^ { x => AttrAlias(Name(x)) }
    | """`[^`]+`""".r ^^ { x => AttrAlias(Name(x.substring(1, x.size - 1))) }
  )

  def relationORsubselect:Parser[RelationORSubselect] = (
      relation ^^ { x => RelationResource(x) }
    | "(" ~ selectORunion ~ ")" ^^ { case "("~s~")" => Subselect(s) }
  )

  def relation:Parser[RDB.RelName] = (
      """[a-zA-Z_]\w*""".r ^^ { x => RDB.RelName(x) }
    | """`[^`]+`""".r ^^ { x => RDB.RelName(x.substring(1, x.size - 1)) }
  )

  def relvar:Parser[RelVar] = (
      """[a-zA-Z_]\w*""".r ^^ { x => RelVar(Name(x)) }
    | """`[^`]+`""".r ^^ { x => RelVar(Name(x)) }
  )

  def tablelist:Parser[(TableList, Set[Expression])] =
    "(?i)FROM".r ~ aliasedjoin ~ rep(innerORouter) ^^
    { case _~aj~l => (TableList(AddOrderedSet(InnerJoin(aj, None) :: l.map((one) => one._1))), 
			   l.foldLeft(Set[Expression]())((all, one) => all ++ one._2)) }

  def innerORouter:Parser[(Join, Set[Expression])] = (
      "(?i)INNER".r ~ "(?i)JOIN".r ~ aliasedjoin ~ opt("(?i)ON".r ~ expression) ^^
      { case _~_~a~o => (InnerJoin(a, None), { if (o.isDefined) Set(o.get._2) else Set[Expression]() } ) }
    | "(?i)LEFT".r ~ "(?i)OUTER".r ~ "(?i)JOIN".r ~ aliasedjoin ~ "(?i)ON".r ~ expression ^^
      { case l~o~j~alijoin~on~expr => (LeftOuterJoin(alijoin, expr), Set[Expression]()) }
  )

  def aliasedjoin:Parser[AliasedResource] =
    relationORsubselect ~ "(?i)AS".r ~ relvar ^^
    { case rel1 ~ _ ~ rel2 => AliasedResource(rel1, rel2) }

  def expression:Parser[Expression] =
    ORexpression ^^ { x => x }

  def ORexpression:Parser[Expression] =
    rep1sep (ANDexpression, "(?i)OR".r) ^^ 
    { xs => if (xs.size > 1) ExprDisjunction(xs.toSet) else xs(0) }

  def ANDexpression:Parser[Expression] =
    rep1sep (relationalexpression, "(?i)AND".r) ^^ 
    { xs => if (xs.size > 1) ExprConjunction(xs.toSet) else xs(0) }

  def relationalexpression:Parser[Expression] = (
      primaryexpression ~ "=" ~ primaryexpression ^^
      { case primaryexpression ~ "=" ~ rvalue => RelationalExpressionEq(primaryexpression, rvalue) }
    | primaryexpression ~ "!=" ~ primaryexpression ^^
      { case primaryexpression ~ "!=" ~ rvalue => RelationalExpressionNe(primaryexpression, rvalue) }
    | primaryexpression ~ "<" ~ primaryexpression ^^
      { case primaryexpression ~ "<" ~ rvalue => RelationalExpressionLt(primaryexpression, rvalue) }
    | primaryexpression ~ ">" ~ primaryexpression ^^
      { case primaryexpression ~ ">" ~ rvalue => RelationalExpressionGt(primaryexpression, rvalue) }
    | primaryexpression ~ "(?i)IS".r ~ "(?i)NULL".r ^^
      { case primaryexpression ~ _ ~ _ => RelationalExpressionNull(primaryexpression) }
    | primaryexpression ~ "(?i)IS".r ~ "(?i)NOT".r ~ "(?i)NULL".r ^^
      { case primaryexpression ~ _ ~ _ ~ _ => RelationalExpressionNotNull(primaryexpression) }
    | primaryexpression ^^
      { case primaryexpression => primaryexpression }
  )

  def primaryexpression:Parser[Expression] = (
      fqattribute ^^ { PrimaryExpressionAttr(_) }
    | float ^^ { i => PrimaryExpressionTyped(RDB.Datatype.FLOAT, Name(i)) }
    | int ^^ { i => PrimaryExpressionTyped(RDB.Datatype.INTEGER, Name(i)) }
    | dquote  ^^ { x => PrimaryExpressionTyped(RDB.Datatype.STRING, Name(x.substring(1, x.size - 1))) }
    | squote  ^^ { x => PrimaryExpressionTyped(RDB.Datatype.STRING, Name(x.substring(1, x.size - 1))) }
    | "(?i)NULL".r ^^ { case _ => ConstNULL() }
    | "(?i)CONCAT".r ~ "(" ~ rep1sep(expression, ",") ~ ")" ^^ { case _~"("~expressions~")" => Concat(expressions) }
    | "(?i)IF".r ~ "(" ~ expression ~ "," ~ expression ~ "," ~ expression ~ ")" ^^ { case _~"("~c~","~p~","~f~")" => IfElse(c, p, f) }
    | "(" ~ expression ~ ")" ^^ { case "("~x~")" => x }
  )

  def value:Parser[RDB.CellValue] = (
      float ^^ { i => RDB.LexicalValue(i) }
    | int ^^ { i => RDB.LexicalValue(i) }
    | dquote  ^^ { x => RDB.LexicalValue(x.substring(1, x.size - 1)) }
    | squote  ^^ { x => RDB.LexicalValue(x.substring(1, x.size - 1)) }
    | "(?i)NULL".r ^^ { case _ => RDB.␀ }
  )

}

case class PrettySql(select:Select) {
  def makePretty():Select = {
    val Select(distinct, projection, tablelist, expression, order, offset, limit) = select
    val nullStripped =
      if (expression.isDefined) {
	val nonNullAttrs = PrettySql.findNonNullRelVarAttrs(expression.get)
	PrettySql.stripNotNulls(expression.get, nonNullAttrs)
      }
      else None
    val XeqXStripped =
      if (nullStripped.isDefined) {
	PrettySql.stripXeqX(nullStripped.get)
      }
      else None
    val strippedTables = tablelist.joins.foldLeft(AddOrderedSet[Join]())((set, join) => set + {
      join match {
	case InnerJoin(AliasedResource(rel, as), optOn) => InnerJoin(AliasedResource({ rel match {
	  case Subselect(s:Select) => Subselect(PrettySql(s).makePretty())
	  case Subselect(Union(disjoints)) => Subselect(Union(disjoints.map(s => PrettySql(s).makePretty())))
	  case r:RelationResource => r
	}}, as), None)
	case LeftOuterJoin(AliasedResource(rel, as), on:Expression) => LeftOuterJoin(AliasedResource({ rel match {
	  case Subselect(s:Select) => Subselect(PrettySql(s).makePretty())
	  case Subselect(Union(disjoints)) => Subselect(Union(disjoints.map(s => PrettySql(s).makePretty())))
	  case r:RelationResource => r
	}}, as), on)
      }})
    Select(distinct, projection, TableList(strippedTables), XeqXStripped, order, offset, limit)
  }
}

object PrettySql {
  def findNonNullRelVarAttrs(expr:Expression):Set[RelVarAttr] = {
    expr match {
      case ExprConjunction(s) => s.foldLeft(Set[RelVarAttr]())((s, e) => s ++ findNonNullRelVarAttrs(e))
      case ExprDisjunction(s) => {
	val l = s.toList
	l.slice(1, l.size).foldLeft(findNonNullRelVarAttrs(l(0)))((s, e) => s & findNonNullRelVarAttrs(e))
      }
      case RelationalExpressionEq(l, r) => findNonNullRelVarAttrs(l) ++ findNonNullRelVarAttrs(r)
      case RelationalExpressionNe(l, r) => findNonNullRelVarAttrs(l) ++ findNonNullRelVarAttrs(r)
      case RelationalExpressionLt(l, r) => findNonNullRelVarAttrs(l) ++ findNonNullRelVarAttrs(r)
      case RelationalExpressionGt(l, r) => findNonNullRelVarAttrs(l) ++ findNonNullRelVarAttrs(r)
      case e:PrimaryExpressionTyped => Set()
      case PrimaryExpressionAttr(a) => Set(a)
      case e:ConstNULL => Set()
      case e:Concat => Set()
      case RelationalExpressionNull(a) => findNonNullRelVarAttrs(a)
      case RelationalExpressionNotNull(a) => Set()
      case IfElse(eef, den, els) => Set()
    }
  }
  def stripNotNulls(expr:Expression, stripMe:Set[RelVarAttr]):Option[Expression] = {
    expr match {
      case ExprConjunction(l) => Some(ExprConjunction({l.foldLeft(Set[Expression]())((s, e) => {
	val e2 = stripNotNulls(e,stripMe)
	if (e2.isDefined) s + e2.get
	else s})}))
      case ExprDisjunction(l) => Some(ExprDisjunction({l.foldLeft(Set[Expression]())((s, e) => {
	val e2 = stripNotNulls(e,stripMe)
	if (e2.isDefined) s + e2.get
	else s})}))
      case e:RelationalExpressionEq => Some(e)
      case e:RelationalExpressionNe => Some(e)
      case e:RelationalExpressionLt => Some(e)
      case e:RelationalExpressionGt => Some(e)
      case e:PrimaryExpressionTyped => Some(e)
      case e:PrimaryExpressionAttr => Some(e)
      case e:ConstNULL => Some(e)
      case e:Concat => Some(e)
      case e:RelationalExpressionNull => Some(e)
      case RelationalExpressionNotNull(PrimaryExpressionAttr(a)) =>
	if (stripMe.contains(a)) None
	else Some(RelationalExpressionNotNull(PrimaryExpressionAttr(a)))
      case e:RelationalExpressionNotNull => Some(e)
      case e:IfElse => Some(e)
    }
  }
  def stripXeqX(expr:Expression):Option[Expression] = {
    expr match {
      case ExprConjunction(l) => Some(ExprConjunction({l.foldLeft(Set[Expression]())((s, e) => {
	val e2 = stripXeqX(e)
	if (e2.isDefined) s + e2.get
	else s})}))
      case ExprDisjunction(l) => Some(ExprDisjunction({l.foldLeft(Set[Expression]())((s, e) => {
	val e2 = stripXeqX(e)
	if (e2.isDefined) s + e2.get
	else s})}))
      case e:RelationalExpressionEq => {
	val RelationalExpressionEq(l, r) = e
	if (l == r) None
	else Some(e)
      }
      case e:RelationalExpressionNe => Some(e)
      case e:RelationalExpressionLt => Some(e)
      case e:RelationalExpressionGt => Some(e)
      case e:PrimaryExpressionTyped => Some(e)
      case e:PrimaryExpressionAttr => Some(e)
      case e:ConstNULL => Some(e)
      case e:Concat => Some(e)
      case e:RelationalExpressionNull => Some(e)
      case e:RelationalExpressionNotNull => Some(e)
      case e:IfElse => Some(e)
    }
  }
  implicit def toPrettySql(select:Select) = PrettySql(select)
}
