package org.w3.rdf.turtle

import org.w3.rdf.{Model => RDFModel, Implicits => RDFImplicits, _}
import scala.util.parsing.combinator._
import java.net.URI

object MyParsers extends RegexParsers {
  val uri = """[a-zA-Z0-9:/#_\.\-\+]+""".r
  val integer = """[0-9]+""".r
  val float = """([0-9]+)?\.([0-9]+)?""".r
  val name = """[a-zA-Z][a-zA-Z0-9_-]*|[a-zA-Z_][a-zA-Z0-9_]+""".r
  var prefixes:Map[String, String] = Map()
  var nextBNode = 1
}

import MyParsers._

trait TurtleModule extends RDFModel {

  // a turtle parser should be instanciated any time you need one
  trait TurtleParser extends JavaTokenParsers {
 
    def toGraph(t:String):Graph = parseAll(turtle, t).get

    def toGraph(file:java.io.File):Graph = {
      val t = scala.io.Source.fromFile(file).getLines.reduceLeft(_+_)
      parseAll(turtle, t).get
    }
 
    def turtle:Parser[Graph] =
      opt(triplesblock) ^^ { case tbOPT => tbOPT.getOrElse(Graph.empty) }
    
    def prefixdecl:Parser[Unit] =
      "@prefix" ~ name ~ ":" ~ qnameORuri ~ "." ^^ { case "@prefix"~pre~":"~u~"." => prefixes += (pre -> { val IRI(iri) = u ; iri }) }
    
    def triplesblock:Parser[Graph] =
      rep(triplepatternOrPrefixOrBase) ^^ {
        case pats => Graph(pats.flatten)
      }
    
    def triplepatternOrPrefixOrBase:Parser[Option[Triple]] = (
        triplepattern ^^ { case p => Some(p) }
      | prefixdecl ^^ { case _ => None }
    )
      
    def triplepattern:Parser[Triple] =
      subject ~ predicate ~ objectt ~ "." ^^ { case s~p~o~"." => Triple(s, p, o) }
  
    def subject:Parser[Subject] = (
        qnameORuri ^^ { case x => SubjectNode(NodeIRI(x)) }
      | bnode ^^ { case x => SubjectNode(NodeBNode(x)) }
    )
      
    def predicate:Parser[Predicate] = (
        qnameORuri ^^ { case x => PredicateIRI(x) }
      | "a" ^^ { x => PredicateIRI(IRI("http://www.w3.org/1999/02/22-rdf-syntax-ns#type")) }
    )
      
    def objectt:Parser[Object] = (
        qnameORuri ^^ { case x => ObjectNode(NodeIRI(x)) }
      | bnode ^^ { case x => ObjectNode(NodeBNode(x)) }
      | literal ^^ { case x => ObjectLiteral(x) }
    )
  
    def qnameORuri:Parser[IRI] = (
        "<"~uri~">" ^^ { case "<"~x~">" => IRI(x) }
      | name~":"~name ^^ {
        case prefix~":"~localName => try {
          IRI(prefixes(prefix) + localName)
        } catch {
          case e:java.util.NoSuchElementException =>
            throw new Exception("unknown prefix " + prefix)
        }
      }
    )
  
    def bnode:Parser[BNode] =
      "_:"~name ^^ { case "_:"~name => BNode(name) }
  
    def literal:Parser[Literal] = (
        stringLiteral~"^^"~qnameORuri ^^ {
        case lit~"^^"~dt => TypedLiteral(lit.substring(1,lit.size - 1), dt match {
          case IRI("http://www.w3.org/2001/XMLSchema#string") => StringDatatype
          case IRI("http://www.w3.org/2001/XMLSchema#integer") => IntegerDatatype
          case IRI("http://www.w3.org/2001/XMLSchema#float") => FloatDatatype
          case IRI("http://www.w3.org/2001/XMLSchema#date") => DateDatatype
          case IRI("http://www.w3.org/2001/XMLSchema#dateTime") => DateTimeDatatype
          case x => error("only programed to deal with string and integer, not " + x)
        })
      }
      | stringLiteral ^^ {
        case lit => PlainLiteral(lit.substring(1,lit.size - 1), None)
      }
      | float ^^ { l => TypedLiteral(l, FloatDatatype) }
      | integer ^^ { l => TypedLiteral(l, IntegerDatatype) }
    )
  }

  case class TurtleSerializer () {
    def graphString (g:Graph):String = {
      g map { t => {
	val Triple(SubjectNode(s), PredicateIRI(p), o) = t
	nodeStr(s) + " " + iriStr(p) + " " + objectStr(o)
      }+".\n" } mkString("")
    }

    def objectStr (n:Object) = {
      n match {
	// case l:ObjectLiteral => {
	//   val x:ObjectLiteral = l
	//   "**ObjectLiteral(" + x + ")**"
	// }
	case ObjectNode(n) => nodeStr(n)
	case ObjectLiteral(l) => literalStr(l)
	// case x => x
      }
    }

    def nodeStr (n:Node) = {
      n match {
	case NodeIRI(i) => iriStr(i)
	case NodeBNode(b) => bnodeStr(b)
      }
    }

    def iriStr (i:IRI) =
      "<" + {
	val IRI(s) = i
	s
      } + ">"

    def bnodeStr (b:BNode) =
      "_:" + {
	val BNode(l) = b
	l
      }
 
    def literalStr (l:Literal) = { l
    }

  }

}


